# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/12_cohort_selector.ipynb.

# %% auto 0
__all__ = ['CohortSelector']

# %% ../nbs/12_cohort_selector.ipynb 3
import re
from typing import List, Any, Dict, Union
import warnings

import numpy as np
import pandas as pd

# %% ../nbs/12_cohort_selector.ipynb 4
from pheno_utils.config import (
    DATASETS_PATH, 
    COHORT, 
    ERROR_ACTION
)
from .meta_loader import MetaLoader

# %% ../nbs/12_cohort_selector.ipynb 5
class CohortSelector:
    """
    Class for selecting a subset of a cohort's data based on a query.

    Args:

        base_path (str, optional): Base path of the datasets. Defaults to DATASETS_PATH.
        cohort (str, optional): Name of the cohort. Defaults to COHORT.
        errors (str, optional): Error action. Defaults to ERROR_ACTION.
        **kwargs: Additional keyword arguments.

    Attributes:

        cohort (str): Name of the cohort.
        base_path (str): Base path of the datasets.
        errors (str): Error action.
        kwargs: Additional keyword arguments.
        ml (MetaLoader): MetaLoader object for loading metadata and data.

    """

    def __init__(
        self,
        base_path: str = DATASETS_PATH,
        cohort: str = COHORT,
        errors: str = ERROR_ACTION,
        **kwargs,
    ) -> None:
        """
        Initialize CohortSelector object.

        Args:

            base_path (str, optional): Base path of the datasets. Defaults to DATASETS_PATH.
            cohort (str, optional): Name of the cohort. Defaults to COHORT.
            errors (str, optional): Error action. Defaults to ERROR_ACTION.
            **kwargs: Additional keyword arguments.

        """
        self.cohort = cohort
        self.base_path = base_path
        self.errors = errors
        self.kwargs = kwargs

        self.ml = MetaLoader(
            base_path=self.base_path, cohort=self.cohort,
            flexible_field_search=False, errors=self.errors,
            **self.kwargs)

    def select(self, query: str) -> pd.DataFrame:
        """
        Select a subset of the cohort's data based on the given query.

        Args:

            query (str): Query string to filter the data.

        Returns:

            pd.DataFrame: Filtered DataFrame based on the query.

        Raises:

            ValueError: If no column names are found in the query.
            ValueError: If column names in the query do not match the column names in the metadata.

        """
        column_names = re.findall(r'([a-zA-Z][a-zA-Z0-9_]*)\b', query)
        if not column_names:
            raise ValueError('No column names found in query')

        test_cols = self.ml.get(column_names)    
        missing_cols = [col for col in column_names
                        if col not in test_cols.columns.str.split('/').str[1]]
        if len(missing_cols):
            raise ValueError(f'Column names {missing_cols} in query do not match column names in metadata')

        df = self.ml.load(column_names)

        return df.query(query)




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/13_questionnaire_handler.ipynb.

# %% auto 0
__all__ = ['convert_to_string', 'tranform_answers']

# %% ../nbs/13_questionnaire_handler.ipynb 3
import pandas as pd
import numpy as np

# %% ../nbs/13_questionnaire_handler.ipynb 4
# Here is a function that can convert a column to string, removing '.0' from floats

def convert_to_string(x):
    return str(int(x)) if isinstance(x, float) and x.is_integer() else str(x)

# Usage:
# df['column_name'] = convert_to_string(df['column_name'])


def tranform_answers(tab_field_name: str, orig_answer: pd.Series, transform_from: str, transform_to: str, 
                     dict_df: pd.DataFrame, mapping_df: pd.DataFrame) -> pd.Series:
    code_from = transform_from.lower()
    code_to = transform_to.lower()
    assert code_from in ['hebrew', 'english', 'coding']
    assert code_to in ['hebrew', 'english', 'coding']
    
    # the index of the dict_df is the tabular_field_name
    if isinstance(dict_df.loc[tab_field_name]['data_coding'], pd.Series):
        code_string = convert_to_string(dict_df.loc[tab_field_name]['data_coding'][0] )
    else: 
        code_string = convert_to_string(dict_df.loc[tab_field_name]['data_coding'])
    code_df = mapping_df[mapping_df['code_number'] == code_string].copy()
    coding = dict(zip(code_df[code_from].astype(int).astype(str), code_df[code_to]))
    return orig_answer.astype(str).replace(coding)




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_date_plots.ipynb.

# %% auto 0
__all__ = ['dates_dist_plot']

# %% ../nbs/04_date_plots.ipynb 3
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Optional, List, Tuple
from pandas._typing import Axes

# %% ../nbs/04_date_plots.ipynb 4
from .config import *

# %% ../nbs/04_date_plots.ipynb 5
def dates_dist_plot(df: pd.DataFrame, col: str, sampling_period: str = "W-MON", ax: Optional[Axes] = None,
                    date_col: str = 'collection_date', ylim: Optional[Tuple[float, float]] = None,
                    quantiles: Optional[List[Tuple[float, str]]] = None) -> None:
    """
    Creates a scatter plot of data points and their statistics based on a specified sampling period.

    Args:
        df (pd.DataFrame): The input DataFrame containing the data.
        col (str): The column name in the DataFrame to plot.
        sampling_period (str, optional): The frequency to resample the data. Defaults to 'W-MON'.
        ax (Optional[Axes], optional): A matplotlib axes object to plot on. Defaults to None.
        date_col (str, optional): The name of the date column in the DataFrame. Defaults to 'collection_date'.
        ylim (Optional[Tuple[float, float]], optional): A tuple defining the y-axis limits. Defaults to None.
        quantiles (Optional[List[Tuple[float, str]]], optional): A list of tuples containing quantiles and their
            labels. Defaults to [(0.1, "10%"), (0.9, "90%")].
    """
    if quantiles is None:
        quantiles = [(0.1, "10%"), (0.9, "90%")]

    df = df.reset_index().set_index(date_col)

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(14, 8))

    # All data scatterplot
    ax.scatter(df.index.get_level_values(date_col), df[col].values, s=40, alpha=0.2,
               marker="o", facecolors="none", linewidths=1, color="k")

    # Define statistics and their styles
    stats = [
        ('quantile', q, f"{label} quantile", "-", 3, 0.4) for q, label in quantiles
    ] + [('median', None, f"Median", "-", 5, 0.5)]

    # Plot sampling period statistics
    for stat, param, label, ls, lw, alpha in stats:
        stat_line = df[col].resample(sampling_period)
        stat_line = stat_line.quantile(param) if stat == 'quantile' else stat_line.median()
        ax.plot(stat_line.index, stat_line.values, label=label, ls=ls, lw=lw, alpha=alpha, color="b")

    # Plot all sample stats
    hline = df[col].median()
    ax.axhline(hline, color="r", label="All samples Median", ls="--", lw=3, alpha=0.5)

    # Prettify the plot
    ax.legend()
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.set_xlabel(date_col, fontsize=16)
    ax.set_ylabel(col, fontsize=16)
    if ylim is not None:
        ax.set_ylim(ylim)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_pheno_loader.ipynb.

# %% auto 0
__all__ = ['PhenoLoader']

# %% ../nbs/05_pheno_loader.ipynb 3
from glob import glob
import traceback

import os
import re
from typing import List, Any, Dict, Union
import warnings
import logging

import numpy as np
import pandas as pd

# %% ../nbs/05_pheno_loader.ipynb 4
from pheno_utils.config import (
    DATASETS_PATH,
    COHORT, 
    EVENTS_DATASET, 
    ERROR_ACTION, 
    BULK_DATA_PATH
    )
from .basic_analysis import custom_describe
from .basic_plots import show_fundus

# %% ../nbs/05_pheno_loader.ipynb 5
class PhenoLoader:
    """
    Class to load multiple tables from a dataset and allows to easily access
    their fields.

    Args:
    
        dataset (str): The name of the dataset to load.
        base_path (str, optional): The base path where the data is stored. Defaults to DATASETS_PATH.
        cohort (str, optional): The name of the cohort within the dataset. Defaults to COHORT.
        age_sex_dataset (str, optional): The name of the dataset to use for computing age and sex. Defaults to EVENTS_DATASET.
        skip_dfs (list, optional): A list of tables (or substrings that match to tables) to skip when loading the data. Defaults to [].
        unique_index (bool, optional): Whether to ensure the index of the data is unique. Defaults to False.
        valid_dates (bool, optional): Whether to ensure that all timestamps in the data are valid dates. Defaults to False.
        valid_stage (bool, optional): Whether to ensure that all research stages in the data are valid. Defaults to False.
        flexible_field_search (bool, optional): Whether to allow regex field search. Defaults to False.
        errors (str, optional): Whether to raise an error or issue a warning if missing data is encountered.
            Possible values are 'raise', 'warn' and 'ignore'. Defaults to 'raise'.

    Attributes:
    
        dict (pd.DataFrame): The data dictionary for the dataset, containing information about each field.
        dfs (dict): A dictionary of dataframes, one for each table in the dataset.
        fields (list): A list of all fields in the dataset.
        dataset (str): The name of the dataset being used.
        cohort (str): The name of the cohort being used.
        base_path (str): The base path where the data is stored.
        dataset_path (str): The full path to the dataset being used.
        age_sex_dataset (str): The name of the dataset being used to compute age and sex.
        skip_dfs (list): A list of tables to skip when loading the data.
        unique_index (bool): Whether to ensure the index of the data is unique.
        valid_dates (bool): Whether to ensure that all timestamps in the data are valid dates.
        valid_stage (bool): Whether to ensure that all research stages in the data are valid.
        flexible_field_search (bool): Whether to allow regex field search.
        errors (str): Whether to raise an error or issue a warning if missing data is encountered.
    """

    def __init__(
        self,
        dataset: str,
        base_path: str = DATASETS_PATH,
        cohort: str = COHORT,
        age_sex_dataset: str = EVENTS_DATASET,
        skip_dfs: List[str] = [],
        unique_index: bool = False,
        valid_dates: bool = False,
        valid_stage: bool = False,
        flexible_field_search: bool = False,
        errors: str = ERROR_ACTION,
        read_parquet_kwargs: Dict[str, Any] = {}
    ) -> None:
        self.dataset = dataset
        self.cohort = cohort
        self.base_path = base_path
        self.dataset_path = self.__get_dataset_path__(self.dataset)
        if self.dataset not in [age_sex_dataset, 'population']:
            self.age_sex_dataset = age_sex_dataset
        else:
            self.age_sex_dataset = None
        self.skip_dfs = skip_dfs
        self.unique_index = unique_index
        self.valid_dates = valid_dates
        self.valid_stage = valid_stage
        self.flexible_field_search = flexible_field_search
        self.errors = errors
        self.read_parquet_kwargs = read_parquet_kwargs

        self.__load_dictionary__()
        self.__load_dataframes__()
        if self.age_sex_dataset is not None:
            self.__load_age_sex__()

    def load_sample_data(
        self,
        field_name: str,
        participant_id: Union[int, List[int]],
        research_stage: Union[None, str, List[str]] = None,
        array_index: Union[None, int, List[int]] = None,
        load_func: callable = pd.read_parquet,
        concat: bool = True,
        pivot=None, **kwargs
    ) -> Union[pd.DataFrame, None]:
        """
        Load time series or bulk data for sample(s).

        Args:
            field_name (str): The name of the field to load.
            participant_id (str or list): The participant ID or IDs to load data for.
            research_stage (str or list, optional): The research stage or stages to load data for.
            array_index (int or list, optional): The array index or indices to load data for.
            load_func (callable, optional): The function to use to load the data. Defaults to pd.read_parquet
            concat (bool, optional): Whether to concatenate the data into a single DataFrame. Automatically ignored if data is not a DataFrame. Defaults to True.
            pivot (str, optional): The name of the field to pivot the data on (if DataFrame). Defaults to None.
        """
        query_str = 'participant_id in @participant_id'
        if not isinstance(participant_id, list):
            participant_id = [participant_id]
        if research_stage is not None:
            if not isinstance(research_stage, list):
                research_stage = [research_stage]
            query_str += ' and research_stage in @research_stage'
        if array_index is not None:
            if not isinstance(array_index, list):
                array_index = [array_index]
            query_str += ' and array_index in @array_index'

        sample = self[[field_name] + ['participant_id']].query(query_str)
        col = sample.columns[0]  # can be different from field_name is a parent_dataframe is implied
        sample = sample.astype({col: str})  
        missing_participants = np.setdiff1d(participant_id, sample['participant_id'].unique())
        sample = sample.loc[:, col]
        

        if len(missing_participants):
            if self.errors == 'raise':
                raise ValueError(f'Missing samples: {missing_participants}')
            elif self.errors == 'warn':
                warnings.warn(f'Missing samples: {missing_participants}')
            if len(sample) == 0:
                return None

        # Load data
        data = []
        for p in sample.unique():
            try:
                data.append(load_func(p, **kwargs))
                if isinstance(data[-1], pd.DataFrame):
                    data[-1].sort_index(inplace=True)
            except Exception as e:
                if self.errors == 'raise':
                    raise e
                elif self.errors == 'warn':
                    warnings.warn(f'Error loading {p}: {e}')

        # Format the final result
        if concat and isinstance(data[0], pd.DataFrame):
            data = pd.concat(data, axis=0)
        if pivot is not None and isinstance(data, pd.DataFrame):
            if pivot in data.index.names:
                data = data.reset_index(pivot)
            data = data.pivot(columns=pivot)

        return data

    def __repr__(self):
        """
        Return string representation of object

        Returns:
            str: String representation of object
        """
        return self.__str__()

    def __str__(self):
        """
        Return string representation of object

        Returns:
            str: String representation of object
        """
        return f'DataLoader for {self.dataset} with' +\
            f'\n{len(self.fields)} fields\n{len(self.dfs)} tables: {list(self.dfs.keys())}'

    def __getitem__(self, fields: Union[str,List[str]]):
        """
        Return data for the specified fields from all tables

        Args:
            fields (Union[str, List[str]]): Fields to return

        Returns:
            pd.DataFrame: Data for the specified fields from all tables
        """
        return self.get(fields)

    def get(self, fields: Union[str,List[str]], flexible: bool=None):
        """
        Return data for the specified fields from all tables

        Args:
            fields (List[str]): Fields to return
            flexible (bool, optional): Whether to use fuzzy matching to find fields. Defaults to None, which uses the DataLoader's flexible_field_search attribute.

        Returns:
            pd.DataFrame: Data for the specified fields from all tables
        """
        if flexible is None:
            flexible = self.flexible_field_search
        if isinstance(fields, str):
            fields = [fields]

        # check whether any field points to a parent_dataframe
        # has_parent = self.dict.loc[self.dict.index.isin(fields), 'parent_dataframe'].dropna()
        seen_fields = set()
        parent_dict = self.dict.loc[self.dict.index.isin(fields), 'parent_dataframe'].dropna().to_dict()
        fields = [parent_dict.get(field, field) for field in fields]
        fields = [field for field in fields if field not in seen_fields and not seen_fields.add(field)]
        # fields += has_parent.unique().tolist()
        flexi_fields = list()

        data = pd.DataFrame()
        for table_name, df in self.dfs.items():
            if 'mapping' in table_name:
                continue
            if flexible:
                # use fuzzy matching including regex to find fields
                fields_in_col = np.unique([col for f in fields for col in df.columns if re.search(f, col)])
                flexi_fields += fields_in_col.tolist()
            else:
                fields_in_col = df.columns.intersection(fields).difference(data.columns)
            if len(fields_in_col):
                data = self.__concat__(data, df[fields_in_col])

            fields_in_index = np.setdiff1d(np.intersect1d(df.index.names, fields), data.columns)
            for field in fields_in_index:
                data = self.__concat__(
                    data,
                    pd.DataFrame(df.index.get_level_values(field), index=df.index))

        if len(data):
            data = data.loc[:, ~data.columns.duplicated()]

        not_found = np.setdiff1d(fields, data.columns)
        if len(not_found) and not flexible:
            if self.errors == 'raise':
                raise KeyError(f'Fields not found: {not_found}')
            elif self.errors == 'warn':
                warnings.warn(f'Fields not found: {not_found}')
        
        data = self.replace_bulk_data_path(data, fields)
        
        cols_order = [field for field in fields if field in data.columns]
        cols_order += [field for field in flexi_fields if (field in data.columns and field not in fields)]

        return data[cols_order]
    

    def replace_bulk_data_path(self, data, fields):
        bulk_fields = self.dict.loc[self.dict.index.isin(fields)].query('item_type == "Bulk"')
        cols = [col for col in bulk_fields.index.to_list() if col in data.columns] 
        dataset_bulk_data_path = {k:v.format(dataset=self.dataset) for k, v in BULK_DATA_PATH.items()}
        category_cols = self.dict.loc[self.dict.index.isin(fields)].query('pandas_dtype == "category"').index
    
        for col in category_cols: 
            data[col] = data[col].astype(str)

        data[cols] = data[cols].fillna('nan').replace(dataset_bulk_data_path, regex=True)
        for col in category_cols: 
            data[col] = data[col].astype('category')
            
        return data



    def __concat__(self, df1, df2):
        if df1.empty:
            return df2
        if df2.empty:
            return df1
        return df1.join(df2, how='outer')

    def __load_age_sex__(self) -> None:
        """
        Add sex and compute age from birth date.
        """
        age_path = os.path.join(self.__get_dataset_path__(self.age_sex_dataset), 'events.parquet')
        align_df = self.dfs[list(self.dfs)[0]]

        if ('research_stage' in align_df.columns) or ('research_stage' in align_df.index.names):
            try:
                age_df = pd.read_parquet(age_path)
                self.dfs['age_sex'] = align_df.join(
                    age_df[['age_at_research_stage', 'sex']].droplevel('array_index'))\
                    .rename(columns={'age_at_research_stage': 'age'})[['age', 'sex']]

            except Exception as e:
                if self.errors == 'raise':
                    raise(e)
                elif self.errors == 'warn':
                    warnings.warn(f'Error joining research_stage: {e}')
                self.dfs['age_sex'] = pd.DataFrame(index=align_df.index).assign(age=np.nan, sex=np.nan)

        else:
            # init an empty df
            self.dfs['age_sex'] = pd.DataFrame(index=align_df.index).assign(age=np.nan, sex=np.nan)

        self.fields += ['age', 'sex']
        ind = self.dfs['age_sex'].isnull().any(axis=1)
        if not ind.any():  # no missing values
            return

        # fill in missing values by computing age from birth date
        try:
            date_cols = np.array(['collection_date', 'collection_timestamp', 'sequencing_date'])
            date = date_cols[np.isin(date_cols, align_df.columns)][0]  # prefer first match
        except Exception as e:
            if self.errors == 'raise':
                raise(e)
            elif self.errors == 'warn':
                warnings.warn(f'No date field found')
            return

        try:
            ind &= align_df[date].notnull()
        except Exception as e:
            if self.errors == 'raise':
                raise(e)
            if self.errors == 'warn':
                warnings.warn(f'Error checking date field: {e}')
            return
        if not ind.any():
            return

        age_df = pd.read_parquet(age_path.replace('events', 'population'))

        # trying a workaround for a pandas deprecation warning
        age_sex = self.dfs['age_sex']
        try:
            age_df['birth_date'] = pd.to_datetime(
                age_df['year_of_birth'].astype(str) + '-' + age_df['month_of_birth'].astype(str))

            missing_age_sex = align_df.loc[ind, [date]].join(age_df[['sex', 'birth_date']])\
                .assign(age=lambda x: ((x[date] - x['birth_date']).dt.days / 365.25).round(1))\
                [['age', 'sex']]
            age_sex = age_sex.join(missing_age_sex, rsuffix='_miss')

        except Exception as e:
            if self.errors == 'raise':
                print("Exception occurred:\n", traceback.format_exc())
                raise(e)
            elif self.errors == 'warn':
                warnings.warn(f'Error joining on {date}: {e}')

            age_sex = age_sex.join(age_df[['sex']], rsuffix='_miss').assign(age_miss=np.nan)

        age_sex['age'] = age_sex['age'].fillna(age_sex['age_miss'])
        age_sex['sex'] = age_sex['sex'].fillna(age_sex['sex_miss'])
        self.dfs['age_sex'] = age_sex[['age', 'sex']]

    def __load_dataframes__(self) -> None:
        """
        Load all tables in the dataset dictionary.
        """
        self.dfs = {}
        self.fields = set()
        for relative_location in self.dict['relative_location'].dropna().unique(): 
            parquet_name = relative_location.split(os.sep)[-1]
            internal_location = os.sep.join(relative_location.split(os.sep)[1:])
            
            
            if any([pattern in relative_location for pattern in self.skip_dfs]):
                print(f'Skipping {relative_location}')
                continue
            df = self.__load_one_dataframe__(internal_location)
            if df is None:
                continue
            table_name = parquet_name.split('.')[0]
            self.dfs[table_name] = df
            if not df.index.is_unique:
                print('Warning: index is not unique for', table_name)
            self.fields |= set(self.dfs[table_name].columns.tolist())
        self.fields = sorted(list(self.fields))

    def __load_one_dataframe__(self, relative_location: str) -> pd.DataFrame:
        """
        Load one dataframe.

        Args:
            relative_location (str): the location of the dataframe

        Returns:
            pd.DataFrame: the loaded dataframe
        """
    
        df_path = os.path.join(self.dataset_path, relative_location)
        
        try:
            data =  pd.read_parquet(df_path, **self.read_parquet_kwargs)
        except Exception as err:
            if self.errors == 'raise':
                warnings.warn(f'Error loading {df_path}:\n{err}')
                raise err
            if self.errors == 'warn':
                warnings.warn(f'Error loading {df_path}:\n{err}')
            return None

        # set the order of columns according to the dictionary
        dict_columns = self.dict.index.intersection(data.columns)
        other_columns = data.columns.difference(self.dict.index)
        assert (len(dict_columns) + len(other_columns)) == len(data.columns), "something isn't right"
        data = data[dict_columns.tolist() + other_columns.tolist()]

        before = len(data)
        if self.unique_index:
            data = data.loc[~data.index.duplicated()]
        if self.valid_dates:
            data = data.loc[data.select_dtypes(include=['datetime64[ns]']).notnull().any(axis=1)]
        if self.valid_stage:
            data = data.loc[data.index.get_level_values('research_stage').notnull()]
        after = len(data)
        if before > after:
            print(f'Filtered {before - after} rows')

        return data

    def __load_dictionary__(self) -> None:
        """
        Load dataset dictionary.
        """
        self.dict = pd.read_csv(self.__get_dictionary_file_path__(self.dataset))\
            .set_index('tabular_field_name')
        self.fields = self.dict.index.tolist()

    def __get_file_path__(self, dataset: str, extension: str) -> str:
        """
        Get the file path for a dataset and an extension.

        Args:
            dataset (str): the name of the dataset
            extension (str): the extension of the file

        Returns:
            str: the path to the file
        """
        path = os.path.join(self.dataset_path, '*.' + extension)
        if path.startswith('s3://'):
           return path
        return glob(path)[0]

    def __get_dictionary_file_path__(self, dataset: str) -> str:
        """
        Get the file path for data dictionary.

        Args:
            dataset (str): the name of the dataset

        Returns:
            str: the path to the file
        """
        path = os.path.join(self.dataset_path, f'{dataset}_data_dictionary.csv')
        if path.startswith('s3://'):
           return path
        return glob(path)[0]

    def __get_dataset_path__(self, dataset):
        """
        Get the dataset path.
        
        Args:
            dataset (str): the name of the dataset

        Returns:
            str: the path to the dataset
        """
        if self.cohort is not None:
            return os.path.join(self.base_path, dataset, self.cohort)
        return os.path.join(self.base_path, dataset)

    def describe_field(self, fields: Union[str,List[str]], return_summary: bool=False):
        """
        Display a summary dataframe for the specified fields from all tables

        Args:
            fields (List[str]): Fields to return
            return_summary (Bool): whether to return the summary dataframe
        
        Returns:
            pd.DataFrame: Data for the specified fields from all tables
        """
        if isinstance(fields, str):
            fields = [fields]
        
        summary_df = pd.concat([self.dict.loc[fields,:].T,
                                custom_describe(self[fields])])
        display(summary_df)
        if return_summary:
            return summary_df



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_config.ipynb.

# %% auto 0
__all__ = ['REF_COLOR', 'FEMALE_COLOR', 'MALE_COLOR', 'ALL_COLOR', 'GLUC_COLOR', 'FOOD_COLOR', 'DATASETS_PATH', 'COHORT',
           'EVENTS_DATASET', 'ERROR_ACTION', 'CONFIG_FILES', 'BULK_DATA_PATH', 'config_found', 'copy_tre_config',
           'generate_synthetic_data', 'generate_synthetic_data_like']

# %% ../nbs/00_config.ipynb 3
import os
import json
import numpy as np
import pandas as pd
import shutil

# %% ../nbs/00_config.ipynb 4
REF_COLOR = "k"
FEMALE_COLOR = "C1"
MALE_COLOR = "C0"
ALL_COLOR = "C5"

GLUC_COLOR = "C0"
FOOD_COLOR = "C1"

DATASETS_PATH = '/home/ec2-user/studies/hpp_datasets/'
COHORT = None
EVENTS_DATASET = 'events'
ERROR_ACTION = 'raise'
CONFIG_FILES = ['.pheno/config.json', '~/.pheno/config.json', '/efs/.pheno/config.json']
BULK_DATA_PATH = {}

config_found = False



# %% ../nbs/00_config.ipynb 5
def copy_tre_config():
    default_config_found = False
    script_path = os.path.dirname(os.path.abspath(__file__))
    
    env_config = os.path.join(script_path, 'config_setup', 'env_config.json')
    with open(env_config, 'r') as openfile:
        env_json = json.load(openfile)
    
    # where am I? 
    config_name = ''
    env = None
    for k, v in env_json.items():
        if os.path.exists(v['ident_path']):
            env = k
            config_name = v['config_name']
            break
        
    absolute_config_path = os.path.join(script_path, 'config_setup', config_name)
    
    if (env is not None) and (os.path.exists(absolute_config_path)):
        default_config_found = True
        if not os.path.exists(os.path.expanduser('~/.pheno')):
            os.makedirs(os.path.expanduser('~/.pheno'))
        
        shutil.copy2(absolute_config_path, os.path.expanduser('~/.pheno/config.json'))
    
    return default_config_found


# %% ../nbs/00_config.ipynb 6
for cf in CONFIG_FILES:
    cf = os.path.expanduser(cf)
    if not os.path.isfile(cf):
        continue
    
    config_found=True
    break


if not config_found: 
    if not copy_tre_config():
        raise ValueError(f'Missing Config file, please read the README file and run config_setup/create_default_config.py')
        
    
for cf in CONFIG_FILES:
    cf = os.path.expanduser(cf)
    if not os.path.isfile(cf):
        continue
    
    f = open(cf)
    config = json.load(f)
    
    if 'DATASETS_PATH' in config:
        DATASETS_PATH = config['DATASETS_PATH']
    if 'BULK_DATA_PATH' in config:
        BULK_DATA_PATH = config['BULK_DATA_PATH']
    if 'EVENTS_DATASET' in config:
        EVENTS_DATASET = config['EVENTS_DATASET']
    if 'COHORT' in config:
        if config['COHORT'] == 0 or config['COHORT']=='None' or config['COHORT']==None :
            COHORT = None
    if 'ERROR_ACTION' in config:
        ERROR_ACTION = config['ERROR_ACTION']
    break


# %% ../nbs/00_config.ipynb 7
def generate_synthetic_data(n: int = 1000) -> pd.DataFrame:
    """
    Generates a sample DataFrame containing age, gender, and value data.

    Args:
        n: The number of rows in the generated DataFrame.

    Returns:
        A pandas DataFrame with columns 'age', 'gender', and 'val'.
    """
    pids = np.arange(n)
    # Set start and end dates
    start_date = pd.Timestamp('2020-01-01')
    end_date = pd.Timestamp('now')
    dates = pd.to_datetime(pd.to_datetime(np.random.uniform(start_date.value, end_date.value, n).astype(np.int64)).date)  
    ages = np.random.uniform(35, 73, size=n)
    genders = np.random.choice([0, 1], size=n)
    vals = np.random.normal(30 + 1 * ages + 40 * genders, 20, size=n)
    
    data = pd.DataFrame(data={"participant_id":pids,"date_of_research_stage": dates,"age_at_research_stage": ages, "sex": genders, "val1": vals}).set_index("participant_id")
    data["val2"] = data["val1"]*0.3 + 0.5*np.random.normal(0,50) + 0.2*10*data["sex"]
    return data

# %% ../nbs/00_config.ipynb 8
def generate_synthetic_data_like(df: pd.DataFrame, n: int = 1000, random_seed: int = 42) -> pd.DataFrame:
    """
    Generate a sample DataFrame containing the same columns as `df`, but with random data.

    Args:
    
        df: The DataFrame whose columns should be used.
        n: The number of rows in the generated DataFrame.

    Returns:
        A pandas DataFrame with the same columns as `df`.
    """
    np.random.seed(random_seed)
    pids = np.arange(n)
    if n > len(df):
        replace = True
    else:
        replace = False

    null = df.reset_index().apply(lambda x: x.sample(frac=1).values)\
        .sample(n=n, replace=replace).assign(participant_id=pids)\
        .set_index(df.index.names)

    def is_path_string(x):
        return isinstance(x, str) and (x.count('/') > 1)

    # handle specific columns
    null.loc[:, null.applymap(is_path_string).mean() > 0.5] = '/path/to/file'
    if ('collection_timestamp' in null.columns) and ('collection_date' in null.columns):
        null['collection_date'] = null['collection_timestamp'].dt.date

    return null



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_basic_plots.ipynb.

# %% auto 0
__all__ = ['data_histplot', 'data_ecdfplot', 'hist_ecdf_plots', 'plot_stats', 'plot_hist_stats', 'plot_data_collection',
           'show_fundus']

# %% ../nbs/01_basic_plots.ipynb 3
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Optional
from PIL import Image
from smart_open import open

# %% ../nbs/01_basic_plots.ipynb 4
from pheno_utils.config import (
    ALL_COLOR,
    FEMALE_COLOR, 
    MALE_COLOR, 
    generate_synthetic_data
    )

# %% ../nbs/01_basic_plots.ipynb 5
def data_histplot(data: pd.DataFrame, col: str, feature_str: Optional[str] = None,
                  gender_col: str = "sex", plot_both_genders: bool=True, ax=None) -> None:
    """
    Plots a histogram from a DataFrame for a specific column.

    Args:
        data (pd.DataFrame): The DataFrame containing the data to plot.
        col (str): The name of the column to plot.
        feature_str (Optional[str], optional): The name of the feature to plot. Defaults to None.
        gender_col (str, optional): The name of the column containing gender information. Defaults to "sex".
        plot_both_genders (bool, optional): Whether to plot both genders or just one. Defaults to True.
        ax ([type], optional): The axis to plot on. Defaults to None.
    """
    try:
        assert col in data.columns, f"Column {col} not found in data"
        if plot_both_genders:
            assert gender_col in data.columns, f"Gender column {gender_col} not found in data"
    except AssertionError as e:
        print(e)
        return

    if feature_str is None:
        feature_str = col

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(6, 4))

    # Plot histogram(s)
    if plot_both_genders:
        idx_male = data[gender_col] == 1
        idx_female = data[gender_col] == 0
        sns.histplot(data.loc[idx_female, col], ax=ax, color=FEMALE_COLOR,
                    label=f'females (N={(data.loc[idx_female, col].count()):,})', alpha=0.5)
        sns.histplot(data.loc[idx_male, col], ax=ax, color=MALE_COLOR,
                    label=f'males (N={(data.loc[idx_male, col].count()):,})', alpha=0.5)
        
    else:
        sns.histplot(data.loc[:, col], ax=ax, color=ALL_COLOR,
                    label=f'N={(data.loc[:, col].count()):,}', alpha=0.5)

    ax.set_xlabel(feature_str, fontsize=16)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.legend()

# %% ../nbs/01_basic_plots.ipynb 9
def data_ecdfplot(data: pd.DataFrame, col: str, feature_str: Optional[str] = None,
                  gender_col: str = "sex", plot_both_genders: bool=True, ax=None) -> None:
    """
    Plots an empirical cumulative distribution function (ECDF) from a DataFrame for a specific column.

    Args:
        data (pd.DataFrame): The DataFrame containing the data to plot.
        col (str): The name of the column to plot.
        feature_str (Optional[str], optional): The name of the feature to plot. Defaults to None.
        gender_col (str, optional): The name of the column containing gender information. Defaults to "sex".
        plot_both_genders (bool, optional): Whether to plot both genders or just one. Defaults to True.
        ax ([type], optional): The axis to plot on. Defaults to None.
    """
    try:
        assert col in data.columns, f"Column {col} not found in data"
        if plot_both_genders:
            assert gender_col in data.columns, f"Gender column {gender_col} not found in data"
    except AssertionError as e:
        print(e)
        return

    if feature_str is None:
        feature_str = col

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(6, 4))

    # Plot ecdf(s)
    if plot_both_genders:
        idx_male = data[gender_col] == 1
        idx_female = data[gender_col] == 0
        sns.ecdfplot(data.loc[idx_female, col], ax=ax, color=FEMALE_COLOR,
                    label=f'females (N={(data.loc[idx_female, col].count()):,})', alpha=0.5, lw=4)
        sns.ecdfplot(data.loc[idx_male, col], ax=ax, color=MALE_COLOR,
                    label=f'males (N={(data.loc[idx_male, col].count()):,})', alpha=0.5, lw=4)
        
    else:
        sns.ecdfplot(data.loc[:, col], ax=ax, color=ALL_COLOR,
                    label=f'N={(data.loc[:, col].count()):,}', alpha=0.5, lw=4)

    ax.set_xlabel(feature_str, fontsize=16)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.legend()

# %% ../nbs/01_basic_plots.ipynb 12
def hist_ecdf_plots(data: pd.DataFrame, col: str, feature_str: Optional[str] = None,
                  gender_col: str = "sex", plot_both_genders: bool=True) -> None:
    """
    Plots histograms and empirical cumulative distribution functions (ECDFs) from a DataFrame
    for a specific column.

    Args:
        data: The input DataFrame containing the data to plot.
        col: The column name to plot.
        feature_str: The title of the plot. If not provided, the column name will be used.
        gender_col: The column name indicating sex (default is 'sex' - female:0; male:1).
        plot_both_genders (bool, optional): Whether to plot both genders or just one. Defaults to True.

    Returns:
        None
    """
    try:
        assert col in data.columns, f"Column {col} not found in data"
        if plot_both_genders:
            assert gender_col in data.columns, f"Gender column {gender_col} not found in data"
    except AssertionError as e:
        print(e)
        return

    if feature_str is None:
        feature_str = col

    fig, axes = plt.subplots(1, 2, figsize=(10, 4))

    # Plot histograms
    ax = axes[0]
    data_histplot(data=data, col=col, feature_str=feature_str,
                  gender_col=gender_col, plot_both_genders=plot_both_genders, ax=ax)
    # Plot ECDFs
    ax = axes[1]
    data_ecdfplot(data=data, col=col, feature_str=feature_str,
                  gender_col=gender_col, plot_both_genders=plot_both_genders, ax=ax)

    fig.tight_layout()
    plt.show()

# %% ../nbs/01_basic_plots.ipynb 14
def plot_stats(data: pd.DataFrame, col: str, ax: plt.Axes, color: str,
               x_position: float = 0.3, label: Optional[str] = "All") -> None:
    """
    Adds a text box to an axis object with summary statistics for a given column in a pandas DataFrame.

    Args:
        data (pd.DataFrame): The pandas DataFrame containing the data to calculate statistics for.
        col (str): The name of the column to calculate statistics for.
        ax (plt.Axes): The axis object to add the text box to.
        color (str): The color of the text box.
        x_position (float, optional): The x position of the text box. Defaults to 0.3.
        label (Optional[str], optional): The label to display in the text box. Defaults to "All".
    """
    # Calculate statistics
    stats = data[col].describe(percentiles=[.10, .90])

    # Rename "50%" to "Median" and "count" to "N"
    stats = stats.rename(index={"50%": "Median", "count": "N"})

    # Ensure "N" is an integer value
    stats["N"] = int(stats["N"])

    # Format statistics for display
    stats_str = f"{label}\n--------------------\n"
    for k, v in stats.items():
        if k == "N":
            stats_str += f"{k}: {int(v)}\n\n"  # Handle "N" separately
        elif np.issubdtype(type(v), np.integer):
            stats_str += f"{k}: {v}\n\n"
        else:
            stats_str += f"{k}: {round(v, 2)}\n\n"

    # Remove trailing newline
    stats_str = stats_str.rstrip("\n")

    # Add textbox to axes
    ax.text(x_position, 0.5, stats_str,
            horizontalalignment='center',
            verticalalignment='center',
            fontsize=9, color='black',
            bbox={'facecolor': color, 'edgecolor': 'black', 'alpha': 0.5, 'pad': 8},
            transform=ax.transAxes)

    # Hide axes
    ax.axis('off')

# %% ../nbs/01_basic_plots.ipynb 15
def plot_hist_stats(data: pd.DataFrame, col: str, feature_str: Optional[str] = None,
                    gender_col: str = "sex", plot_both_genders: bool=True) -> None:
    """
    Plots a histogram of a given column in a pandas DataFrame and adds summary statistics to the plot.

    Args:
        data (pd.DataFrame): The pandas DataFrame containing the data to plot.
        col (str): The name of the column to plot.
        feature_str (Optional[str], optional): A string describing the feature being plotted. Defaults to None.
        gender_col (str, optional): The name of the column containing gender information. Defaults to "sex".
        plot_both_genders (bool, optional): Whether to plot statistics separately for males and females. Defaults to True.
    """
    # Create a figure and an axes
    fig, axes = plt.subplots(1,2, figsize=(10, 4))

    # Plot histogram on first axis
    ax = axes[0]
    data_histplot(data, col, feature_str, gender_col, plot_both_genders, ax=ax)
    ax.get_legend().remove()  # Remove legend from histogram

    # Plot statistics on second axis
    ax = axes[1]
    if plot_both_genders:
        # Separate data by gender
        idx_male = data[gender_col] == 1
        idx_female = data[gender_col] == 0

        # Plot statistics for males and females separately
        plot_stats(data.loc[idx_male], col, ax=ax, color=MALE_COLOR, x_position=0, label="Males")
        plot_stats(data.loc[idx_female], col, ax=ax, color=FEMALE_COLOR, x_position=0.4, label="Females")

        # Plot statistics for all data
        plot_stats(data.loc[:], col, ax=ax, color=ALL_COLOR, x_position=0.8)

    else:
        # Plot statistics for all data
        plot_stats(data, col, ax=ax, color=ALL_COLOR, x_position=0)

# %% ../nbs/01_basic_plots.ipynb 19
def plot_data_collection(data: pd.DataFrame, date_col: str = "collection_date", feature_str: Optional[str] = None, ax: Optional[plt.Axes] = None) -> None:
    """
    Plots a histogram of the specified column in a pandas DataFrame and excludes the last point from the plot.

    Args:
        data (pd.DataFrame): The pandas DataFrame containing the data to plot.
        date_col (str, optional): The name of the column containing the dates. Defaults to "collection_date".
        feature_str (Optional[str], optional): The name of the feature to plot. If None, the name of the date column will be used. Defaults to None.
        ax (Optional[plt.Axes], optional): The axis object to plot on. If None, a new figure and axis will be created. Defaults to None.
    """
    if feature_str is None:
        feature_str = ""

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(6, 3))

    df = data.sort_values(date_col).dropna()

    # Calculate cumulative count of rows
    df['cumulative_count'] = range(1, len(df) + 1)

    # Plot cumulative count over time
    ax.plot(df[date_col], df['cumulative_count'], lw=3, alpha=0.8)

    # Rotate x-axis tick labels by 90 degrees
    plt.xticks(rotation=90)

    # Remove top and right spines
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)

    ax.set_ylabel(f"{feature_str} data collected", fontsize=14)

# %% ../nbs/01_basic_plots.ipynb 21
def show_fundus(fname: str) -> None:
    """
    Display a fundus image from an input file path.
    Args:
        fname (str): The file path to the fundus image.
    """
    fig, ax = plt.subplots(1, 1, figsize=(6, 6))
    img = Image.open(open(fname, 'rb'))
    ax.imshow(img, cmap="gray")
    ax.set_xticks([])
    ax.set_yticks([])
    ax.axis('off')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/09_ecg_analysis.ipynb.

# %% auto 0
__all__ = ['vis_ecg', 'get_hrv_df']

# %% ../nbs/09_ecg_analysis.ipynb 3
from typing import Tuple

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

import neurokit2 as nk

# %% ../nbs/09_ecg_analysis.ipynb 4
def vis_ecg(values_df: pd.DataFrame) -> None:
    """
    Visualize ECG data for 12 leads.

    Args:
        values_df (pd.DataFrame): A DataFrame containing ECG data with 12 columns, one for each lead.

    Returns:
        None: Displays a 3x4 grid of ECG plots for the 12 leads.
    """
    count = 0
    frequency = 1000
    time_data = np.arange(values_df.shape[0]) / frequency
    sns.set(rc={'figure.figsize':(40.7,18.27)})
    fig, axs = plt.subplots(3, 4)
    
    for y in range(0,4):    
        for x in range(0,3):
            col = values_df.columns[count]
            axs[x, y].plot(time_data, values_df[col].values)
            axs[x, y].set_title(col)
            count += 1
            plt.xlabel("time in seconds")
            plt.ylabel("ECG in uV")


# %% ../nbs/09_ecg_analysis.ipynb 5
def get_hrv_df(ECG_df: pd.DataFrame, sr: int = 1000) -> pd.DataFrame:
    """
    Compute the Heart Rate Variability (HRV) metrics for each ECG lead in the input DataFrame.

    Args:
        ECG_df (pd.DataFrame): A DataFrame containing ECG data with one column for each lead.
        sr (int, optional): The sampling rate of the ECG data. Defaults to 1000.

    Returns:
        pd.DataFrame: A DataFrame containing HRV metrics for each ECG lead.
    """
    hrv_list = []
    peaks_list = []
    
    for i in ECG_df.columns:
        peaks, info = nk.ecg_peaks(ECG_df[i], sampling_rate=sr, correct_artifacts=True)
        hrv = nk.hrv_time(peaks, sampling_rate=sr, show=False)
        hrv_list.append(hrv)
        peaks_list.append(peaks)

    hrv_df = pd.concat(hrv_list)
    hrv_df.index = ECG_df.columns
    hrv_df = hrv_df.dropna(axis=1, how='all')
    
    peaks_df = pd.concat(peaks_list, axis=1)
    peaks_df.columns = ECG_df.columns

    return hrv_df




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_age_reference_plots.ipynb.

# %% auto 0
__all__ = ['get_gam_expectiles', 'AgeRefPlot', 'GenderAgeRefPlot']

# %% ../nbs/03_age_reference_plots.ipynb 3
from pheno_utils.config import (
    generate_synthetic_data, 
    FEMALE_COLOR,
    MALE_COLOR,
    ALL_COLOR, 
    REF_COLOR,
    ERROR_ACTION, 
    )

from typing import Dict, List, Callable, Optional, Union, Tuple
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import importlib
from scipy.stats import linregress
from sklearn.linear_model import HuberRegressor
from numpy import ndarray


# %% ../nbs/03_age_reference_plots.ipynb 4
try:
    importlib.import_module("pygam")
except ImportError:
    raise ImportError("The 'pygam' library is not installed. Please install it using 'pip install pygam' before running this script.")

from pygam import ExpectileGAM

# %% ../nbs/03_age_reference_plots.ipynb 5
def get_gam_expectiles(X: ndarray, y: ndarray, expectiles: List[float] = [0.03, 0.1, 0.5, 0.9, 0.97]) -> Tuple[ndarray, Dict[str, ndarray]]:
    """
    Fit Expectile Generalized Additive Models (GAMs) for given expectiles.

    Args:
        X (ndarray): Feature data for the model.
        y (ndarray): Target variable for the model.
        expectiles (List[float]): List of expectiles to fit. Default is [0.03, 0.1, 0.5, 0.9, 0.97].

    Returns:
        Tuple[ndarray, Dict[str, ndarray]]: A tuple containing a grid of X values for prediction and a dictionary with expectiles as keys and their corresponding model predictions as values.
    """

    # Fit the median model first by Cross-Validation
    gam50 = ExpectileGAM(expectile=0.5).gridsearch(X, y, progress=False)

    # Copy the smoothing to the other models
    lam = gam50.lam

    # Generate a grid of X values for prediction
    XX = gam50.generate_X_grid(term=0, n=500)

    # Initialize dictionary to store the GAM expectiles
    gam_expectiles = {}

    # Fit all wanted expectiles and store their predictions
    for expectile in expectiles:
      gam = ExpectileGAM(expectile=expectile, lam=lam).fit(X, y)
      gam_expectiles[expectile] = gam.predict(XX)

    return XX, gam_expectiles

# %% ../nbs/03_age_reference_plots.ipynb 6
class AgeRefPlot:
    def __init__(
        self,
        data: pd.DataFrame,
        val_col: str,
        age_col: str = "age_at_research_stage",
        sex_col: str = "sex",
        sex: Optional[int] = None,
        val_color: Optional[str] = None,
        ax_main: Optional[plt.Axes] = None,
        ax_agehist: Optional[plt.Axes] = None,
        ax_valhist: Optional[plt.Axes] = None,
        age_bins: Optional[np.ndarray] = None,
        val_bins: Optional[np.ndarray] = None,
        linear_fit: bool = True,
        expectiles: Optional[List] = [0.03, 0.1, 0.5, 0.9, 0.97],
        top_disp_perc: float = 99,
        bottom_disp_perc: float = 1,
        robust: bool = True,
        scale: float = 1.,
        transform: Optional[Callable] = None,
        make_fig: bool = True
    ) -> None:
        """
        Initializes the AgeRefPlot class.

        Args:
            data (pd.DataFrame): A pandas DataFrame containing the data.
            val_col (str): The name of the value column in the DataFrame.
            age_col (str): The name of the age column in the DataFrame.
            sex_col (str): The name of the sex column in the DataFrame.
            sex (Optional[int], optional): The sex to filter the data by. 0 for females and 1 for males. Defaults to None.
            val_color (Optional[str], optional): The color to use for the value plot. Defaults to None.
            ax_main (Optional[plt.Axes], optional): The main axis for the plot. Defaults to None.
            ax_agehist (Optional[plt.Axes], optional): The axis for the age histogram. Defaults to None.
            ax_valhist (Optional[plt.Axes], optional): The axis for the value histogram. Defaults to None.
            age_bins (Optional[np.ndarray], optional): The age bins for the histograms. Defaults to None.
            val_bins (Optional[np.ndarray], optional): The value bins for the histograms. Defaults to None.
            linear_fit (bool, optional): Whether to perform a linear fit on the data. Defaults to True.
            expectiles (Optional[List], optional): Whether to calculate and shpe gam expectiles or not. Defaults to [0.03, 0.1, 0.5, 0.9, 0.97].
            top_disp_perc (float, optional): The top percentile to use for display. Defaults to 99.
            bottom_disp_perc (float, optional): The bottom percentile to use for display. Defaults to 1.
            robust (bool, optional): Whether to use a robust regression method (HuberRegressor) instead of ordinary least squares for linear_fit. Defaults to True.
            scale (float, optional): The scaling factor for the value column. Defaults to 1.
            transform (Optional[Callable], optional): The transformation function to apply to the value column. Defaults
            make_fig (bool, optional): Whether to create a new figure if axes are not provided. Defaults to True.
        """
        self.data = data.dropna(subset=[age_col, val_col]).copy()
        self.data = self.data.sort_values(by=age_col)
        if transform is not None:
            self.data[val_col] = transform(self.data[val_col])
        self.data[val_col] = scale * self.data[val_col]
        self.val_col = val_col
        self.age_col = age_col
        self.sex_col = sex_col
        self.transform = transform
        self.scale = scale
        self.sex = sex
        self.linear_fit = linear_fit
        self.expectiles = expectiles
        # Cut data for display only removing outliers
        self.top_disp_perc = top_disp_perc/100
        self.bottom_disp_perc = bottom_disp_perc/100
        quantiles = self.data[[val_col]].quantile([self.bottom_disp_perc, self.top_disp_perc])
        self.disp_data = self.data[(self.data[val_col] > quantiles.loc[self.bottom_disp_perc, val_col]) &
                                   (self.data[val_col] < quantiles.loc[self.top_disp_perc, val_col])].copy()
        self.robust = robust
        self.slope = np.nan
        self.intercept = np.nan

        if val_color is None:
            self.val_color = {0: FEMALE_COLOR, 1: MALE_COLOR}.get(self.sex, ALL_COLOR)
        self.ref_color = REF_COLOR

        if age_bins is None:
            age_bins = np.arange(35, 75, 1)
        self.age_bins = age_bins
        self.min_age = self.age_bins[0]
        self.max_age = self.age_bins[-1]

        self.expectiles_dict = None

        if val_bins is None:
            val_bins = np.linspace(
                self.disp_data[self.val_col].min(), self.disp_data[self.val_col].max(), 50
            )
        self.val_bins = val_bins

        # Prep axes
        if (((ax_main is None) | (ax_agehist is None) | (ax_valhist is None)) & (make_fig==True)):
            ax_dict = plt.figure(
                constrained_layout=True, figsize=(7, 5)
            ).subplot_mosaic(
                """
            AAA.
            CCCE
            CCCE
            CCCE
            CCCE
            """,
                gridspec_kw={"wspace": 0, "hspace": 0},
            )
            self.ax_agehist = ax_dict["A"]
            self.ax_valhist = ax_dict["E"]
            self.ax_main = ax_dict["C"]

        else:
            self.ax_agehist = ax_agehist
            self.ax_valhist = ax_valhist
            self.ax_main = ax_main

    def plot_agehist(self):
        self.ax_agehist.hist(
            self.disp_data[self.age_col],
            bins=self.age_bins,
            facecolor=self.val_color,
            edgecolor="white",
            linewidth=0.5,
            alpha=0.8,
        )

    def plot_valhist(self):
        self.ax_valhist.hist(
            self.disp_data[self.val_col],
            orientation="horizontal",
            bins=self.val_bins,
            facecolor=self.val_color,
            edgecolor="white",
            linewidth=0.5,
            alpha=0.8,
        )

    def plot_scatter(self):
        self.ax_main.scatter(
            self.disp_data[self.age_col],
            self.disp_data[self.val_col],
            s=40,
            alpha=0.4,
            marker="o",
            facecolors="none",
            linewidths=1,
            color=self.val_color,
        )

    def plot_percentiles(self):
        # fetch percentiles if needed by fitting a GAM
        if self.expectiles_dict is None:
            X, y = self.data[[self.age_col]].values, self.data[self.val_col].values
            XX, self.expectiles_dict = get_gam_expectiles(X, y, self.expectiles)

        middle_index = len(self.expectiles) // 2

        # plot expectiles curves
        for perc_str in self.expectiles:
            perc_vector = self.expectiles_dict[perc_str]
            self.ax_main.plot(
                XX,
                perc_vector,
                color=self.ref_color,
                alpha=0.5,
                lw=2,
                ls=":",
            )
            self.ax_main.text(
                x=XX.max(),
                y=perc_vector[-1],
                s=perc_str,
                alpha=0.5,
            )

        # fill between curves
        if len(self.expectiles) > 2:
            self.ax_main.fill_between(
                XX.squeeze(),
                self.expectiles_dict[(self.expectiles[middle_index - 1])],
                self.expectiles_dict[(self.expectiles[middle_index + 1])],
                alpha=0.1,
                color=self.ref_color,
            )
        if len(self.expectiles) > 4:
            self.ax_main.fill_between(
                XX.squeeze(),
                self.expectiles_dict[(self.expectiles[middle_index - 2])],
                self.expectiles_dict[(self.expectiles[middle_index + 2])],
                alpha=0.05,
                color=self.ref_color,
            )

    def calc_linear_fit(self):
        # get coeffs of linear fit
        if self.robust:
            huber = HuberRegressor().fit(
                self.data[[self.age_col]].values, self.data[[self.val_col]].values.ravel()
            )
            self.slope = huber.coef_[0]
            self.intercept = huber.intercept_
            r_value = np.nan
            p_value = np.nan
            std_err = np.nan
        else:
            self.slope, self.intercept, r_value, p_value, std_err = linregress(
                self.data[self.age_col].values, self.data[self.val_col].values
            )
        return self.slope, self.intercept, r_value, p_value, std_err

    def plot_ornaments(self):
        ax_main_str = {0: "Females", 1: "Males"}.get(self.sex, "")
        if self.linear_fit:
            slope, intercept, r_value, p_value, std_err = self.calc_linear_fit()
            fit_str = "y={0:.3f}*age+{1:.3f}".format(slope, intercept)
            ax_main_str = ax_main_str + " " + self.val_col + "\n" + fit_str

        self.ax_main.text(
            0.05,
            0.9,
            ax_main_str,
            color="black",
            fontsize="x-large",
            weight="bold",
            ha="left",
            va="baseline",
            transform=self.ax_main.transAxes,
            # font="Roboto Condensed",
        )

    def plot(self):
        ax = self.ax_main
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)

        ax = self.ax_agehist
        self.ax_main.sharex(self.ax_agehist)
        ax.set_yticks([])
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.tick_params(labelbottom=False) # hides x-tick labels

        ax = self.ax_valhist
        self.ax_main.sharey(self.ax_valhist)
        ax.set_xticks([])
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.tick_params(labelleft=False) # hides y-tick labels


        ax = self.ax_main
        ax.set_xlabel("Age")
        xticks = (np.sort(np.unique(np.floor(self.age_bins / 10) * 10))).astype(int)
        xticks = np.append(xticks, xticks[-1] + 10)
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticks)
        ax.set_xlim(xticks[0], xticks[-1] + 1)
        ax.set_ylabel(self.val_col)

        self.plot_scatter()
        self.plot_percentiles()
        self.plot_ornaments()
        self.plot_agehist()
        self.plot_valhist()


# %% ../nbs/03_age_reference_plots.ipynb 8
class GenderAgeRefPlot(AgeRefPlot):
    def __init__(
        self, 
        data: pd.DataFrame,
        val_col: str,
        age_col: str = "age_at_research_stage",
        sex_col: str = "sex",
        age_bins: Optional[np.ndarray] = None,
        val_bins: Optional[np.ndarray] = None,
        linear_fit: bool = True,
        expectiles: Optional[List] = [0.03, 0.1, 0.5, 0.9, 0.97],
        top_disp_perc: float = 99,
        bottom_disp_perc: float = 1,
        robust: bool = True,
        scale: float = 1.,
        transform: Optional[Callable] = None,
    ) -> None:
        """Initializes the GenderAgeRefPlot class.

        Args:
            data (pd.DataFrame): The input data containing age, value, and gender columns.
            val_col (str): The name of the value column in the data.
            age_col (str): The name of the age column in the DataFrame.
            sex_col (str): The name of the sex column in the DataFrame.
            age_bins (np.ndarray, optional): An array of age bin edges.
            val_bins (np.ndarray, optional): An array of value bin edges.
            linear_fit (bool, optional): Whether to fit a linear regression line. Defaults to True.
            expectiles (Optional[List], optional): Whether to calculate and shpe gam expectiles or not. Defaults to [0.03, 0.1, 0.5, 0.9, 0.97].
            top_disp_perc (float, optional): The top percentile for data display. Defaults to 99.
            bottom_disp_perc (float, optional): The bottom percentile for data display. Defaults to 1.
            robust (bool, optional): Whether to use a robust linear regression. Defaults to True.
            scale (float, optional): The scaling factor for the data. Defaults to 1.
            transform (Callable, optional): An optional function to apply to the data. Defaults to None.
        """
        super().__init__(
            data=data,
            val_col=val_col,
            age_col=age_col,
            sex_col=sex_col,
            age_bins=age_bins,
            val_bins=val_bins,
            linear_fit=linear_fit,
            expectiles=expectiles,
            top_disp_perc=top_disp_perc,
            bottom_disp_perc=bottom_disp_perc,
            robust=robust,
            scale=scale,
            transform=transform,
            make_fig=False
        )

    def plot(self) -> None:
        """Plots the data for both genders in separate panels."""
        ax_dict = plt.figure(constrained_layout=True, figsize=(16, 6)).subplot_mosaic(
            """
            AAA.BBB.
            CCCEDDDF
            CCCEDDDF
            CCCEDDDF
            CCCEDDDF
            """,
            gridspec_kw={"wspace": 0, "hspace": 0},
        )
        self.ax_dict = ax_dict

        self.female_refplot = AgeRefPlot(
            data=self.data[self.data["sex"]==0],
            val_col=self.val_col,
            age_col=self.age_col,
            sex_col=self.sex_col,
            sex=0,
            ax_main=ax_dict["C"],
            ax_agehist=ax_dict["A"],
            ax_valhist=ax_dict["E"],
            age_bins=self.age_bins,
            val_bins=self.val_bins,
            linear_fit=self.linear_fit,
            top_disp_perc=self.top_disp_perc*100,
            bottom_disp_perc=self.bottom_disp_perc*100,
            expectiles=self.expectiles,
            robust=self.robust,
        )
        self.female_refplot.plot()

        self.male_refplot = AgeRefPlot(
            data=self.data[self.data["sex"]==1],
            val_col=self.val_col,
            age_col=self.age_col,
            sex_col=self.sex_col,
            sex=1,
            ax_main=ax_dict["D"],
            ax_agehist=ax_dict["B"],
            ax_valhist=ax_dict["F"],
            age_bins=self.age_bins,
            val_bins=self.val_bins,
            linear_fit=self.linear_fit,
            top_disp_perc=self.top_disp_perc*100,
            bottom_disp_perc=self.bottom_disp_perc*100,
            expectiles=self.expectiles,
            robust=self.robust,
        )
        self.male_refplot.plot()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_sleep_plots.ipynb.

# %% auto 0
__all__ = ['CHANNELS', 'DEFAULT_CHANNELS', 'COLOR_GROUPS', 'ENUMS', 'CHANNEL_LIMS', 'plot_sleep', 'plot_events', 'plot_channels',
           'format_xticks', 'get_legend_colors']

# %% ../nbs/06_sleep_plots.ipynb 3
from typing import Iterable, Optional

import numpy as np
import pandas as pd

import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

# %% ../nbs/06_sleep_plots.ipynb 4
CHANNELS = {
    'actigraph': 'Actigraph',
    'body_position': 'Body Position',
    'heart_rate': 'Heart Rate',
    'heart_rate_raw': 'Heart Rate Raw',
    'pat_infra': 'PAT Infra',
    'pat_amplitude': 'PAT Amplitude',
    'pat_lpf': 'PAT LPF',
    'respiratory_movement': 'Respiratory Mov.',
    'spo2': 'SpO2',
    'snore_db': 'Snore dB',
    'pat_view': 'PAT View',
    'sleep_stage': 'Sleep Stage'
    }

DEFAULT_CHANNELS = ['actigraph', 'pat_infra', 'body_position', 'snore_db', 'heart_rate', 'spo2']

COLOR_GROUPS= {
    'actigraph': ['actigraph', 'sleep_stage'],
    'general': ['body_position'],
    'heart_rate': ['heart_rate'],
    'heart_rate_raw': ['heart_rate_raw'],
    'pat_amplitude': ['pat_amplitude'],
    'pat_infra': ['pat_infra', 'pat_view', 'pat_lpf'],
    'respiratory_movement': ['respiratory_movement', 'snore_db'],
    'spo2': ['spo2'],
    }

ENUMS = {'body_position': [[0, 1, 2, 3, 4, 5], ['N/A', 'Supine', 'Right', 'Left', 'Prone', 'Sit']],
         'sleep_stage': [[11, 12, 13, 14, 15], ['Wake', 'REM', 'Light', 'End', 'Deep']]}
CHANNEL_LIMS = {'spo2': [0, 100]}

# %% ../nbs/06_sleep_plots.ipynb 5
def plot_sleep(events: pd.DataFrame, channels: pd.DataFrame,
               array_index: Optional[int] = None,
               trim_to_events: Optional[bool] = True,
               add_events: Optional[pd.DataFrame] = None,
               event_filter: Optional[Iterable[str]] = None,
               channel_filter: Optional[Iterable[str]] = DEFAULT_CHANNELS,
               event_height: float=2, channel_height: float=0.45, width: float=10, aspect: float=0.2,
               style: str='whitegrid',
               xlim: Iterable[float]=None, **kwargs):
    """
    Plot sleep events and channels data.

    Args:

        events (pd.DataFrame): A pandas dataframe containing sleep events data.
        channels (pd.DataFrame): A pandas dataframe containing raw channels data.
        array_index (int, optional): The index of the array. Defaults to None.
        trim_to_events (bool, optional): Whether to trim the plot to the start and end of the events. Defaults to True.
        add_events (pd.DataFrame, optional): Additional events data to include in the plot. Defaults to None.
        event_filter (Iterable[str], optional): A list of events to include in the plot. Defaults to None.
        channel_filter (Iterable[str], optional): A list of channels to include in the plot. Defaults to DEFAULT_CHANNELS.
        event_height (float, optional): The height of the event plot in inches. Defaults to 2.
        channel_height (float, optional): The height of each channel plot in inches. Defaults to 0.45.
        width (float, optional): The width of the plot in inches. Defaults to 10.
        aspect (float, optional): The aspect ratio of the plot. Defaults to 0.2.
        style (str, optional): The seaborn style to use. Defaults to 'whitegrid'.
        xlim (List[float], optional): The x-axis limits of the plot. Defaults to None.
        **kwargs: Additional arguments to be passed to plot_channels().

    Returns:

        None
    """
    nC = min([len(channel_filter), channels.index.get_level_values('source').nunique()])
    if xlim is not None:
        trim_to_events = True

    fig, ax = plt.subplots(nrows=nC+1, ncols=1, sharex=True, squeeze=False,
                           height_ratios=nC*[channel_height] + [event_height],
                           figsize=(width, width*aspect*(nC*channel_height + event_height)),
                           facecolor='white')

    with sns.axes_style(style):
        try:
            plot_channels(channels, array_index=array_index, y_filter=channel_filter, ax=ax[:-1],
                     **kwargs)
            plot_events(events, array_index=array_index, y_include=event_filter, y_exclude=['Gross Trunc'],
                        ax=ax[-1,0],
                        set_xlim=trim_to_events, add_events=add_events,
                        xlim=xlim)
        except Exception as err:
            print(f'plot_channels failed due to:\n{err}')
            fig.clf()
            plot_events(events, array_index, y_include=event_filter, y_exclude=['Gross Trunc'],
                        set_xlim=trim_to_events, add_events=add_events,
                        xlim=xlim)


def plot_events(events: pd.DataFrame, array_index: Optional[int]=None,
                x_start: str='collection_timestamp', x_end: str='event_end',
                y: str='event', color: str='channel', cmap: str='muted', set_xlim: bool=True,
                xlim: Iterable[float]=None, figsize: Iterable[float]=[10, 4],
                y_include: Optional[Iterable[str]] = None,
                y_exclude: Optional[Iterable[str]] = None,
                ax: plt.Axes=None,
                add_events: Optional[pd.DataFrame] = None,
                rename_channels: dict={'PAT Amplitude': 'PAT', 'PulseRate': 'Heart Rate'},
                rename_events: dict={}):
    """ plot an events timeline for a given participant and array_index """
    # slice to night
    if (array_index is not None) and (('array_index' in events.columns) or ('array_index' in events.index.names)):
        plot_df = events.query('array_index == @array_index').copy()
    else:
        plot_df = events.copy()
    # extract start and end times
    if x_start in plot_df.index.names:
        plot_df = plot_df.reset_index(x_start)
    if x_end in plot_df.index.names:
        plot_df = plot_df.reset_index(x_end)
    # remove timezone for correct matplotlib labeling
    plot_df[x_start] = plot_df[x_start].dt.tz_localize(None)
    plot_df[x_end] = plot_df[x_end].dt.tz_localize(None)

    # filter events
    if y_include is not None:
        plot_df = plot_df.query(f'{y} in {y_include}')
    if y_exclude is not None:
        plot_df = plot_df.query(f'{y} not in {y_exclude}')
    # additional user-provided events (application logging, etc.)
    if add_events is not None:
        tlim = plot_df[x_start].min(), plot_df[x_end].max()
        add_events = add_events.loc[
            (tlim[0] < add_events[x_end]) & (add_events[x_start] < tlim[1])]
        if len(add_events):
            add_events = add_events.set_index(plot_df.index[[0]])
        plot_df = pd.concat([plot_df, add_events[
            plot_df.columns.intersection(add_events.columns)]], axis=0)

    # rename channels and events
    plot_df = plot_df.copy()
    plot_df['channel'] = plot_df['channel'].replace(rename_channels)
    plot_df['event'] = plot_df['event'].replace(rename_events)

    # set x limits
    if xlim is not None:
        if type(xlim[0]) is str:
            xlim = (pd.to_datetime(xlim[0]), xlim[1])
        if type(xlim[0]) is not pd.Timestamp:
            xlim = plot_df.loc[plot_df['start'] < xlim[0], x_start].iloc[-1], xlim[1]
        if type(xlim[1]) is str:
            xlim = (xlim[0], pd.to_datetime(xlim[1]))
        if type(xlim[1]) is not pd.Timestamp:
            xlim = xlim[0], plot_df.loc[plot_df['end'] > xlim[1], x_end].iloc[0]
    else:
        xlim = plot_df[x_start].min(), plot_df[x_end].max()

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    # set colors
    colors = sorted(plot_df[color].unique())
    colors = pd.DataFrame({color: colors, 'color': sns.color_palette(cmap, len(colors))})\
        .set_index(color)['color']

    # plot events
    plot_df = plot_df.assign(diff=lambda x: x[x_end] - x[x_start]).sort_values([color, y])
    labels = []
    legend = []
    for i, (y_label, y) in enumerate(plot_df.groupby(y, sort=False)):
        if len(y) == 0:
            continue
        labels.append(y_label)
        for c, r in y.groupby(color):
            data = r[[x_start, 'diff']]
            if not len(data):
                continue
            h = ax.broken_barh(data.values, (i-0.4,0.8), color=colors[c], alpha=0.7)
            legend.append({'label': c, 'handle': h})

    # format plot
    legend = pd.DataFrame.from_dict(legend).drop_duplicates(subset='label')
    ax.legend(legend['handle'], legend['label'],
              bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0.)

    str_title = ''
    if 'participant_id' in events.index.names:
        str_title += events.index.get_level_values('participant_id')[0].astype(str)
    if array_index is not None:
        str_title += f' / {array_index}'
    plt.suptitle(str_title, fontsize=14, weight='bold')
    ax.set_yticks(np.arange(len(labels)), labels)
    plt.tight_layout()
    ax.set_xlabel('Time')
    if set_xlim:
        ax.set_xlim(*xlim)
    format_xticks(ax)

    return ax


def plot_channels(channels: pd.DataFrame, array_index: Optional[int]=None,
             y_filter: Optional[Iterable[str]]=None, ax: plt.Axes=None,
             discrete_events: Optional[Iterable[str]]=['sleep_stage', 'body_position'],
             time_col='collection_timestamp', height=1.5, resample='1s', cmap='muted',
             rename_channels=CHANNELS, **kwargs):
    """ plot channels data for a given participant and array_index """
    # set colors
    colors = get_legend_colors(cmap).explode('source')
    colors['source'] = pd.Categorical(colors['source'])
    colors = colors.set_index('source')

    # filter data
    if (array_index is not None) and (('array_index' in channels.columns) or ('array_index' in channels.index.names)):
        data = channels.query('array_index == @array_index').copy()
    else:
        data = channels.copy()
    # extract time and channel name
    if time_col in channels.index.names:
        data = data.reset_index(time_col)
    if 'source' not in data.index.names:
        data = data.set_index('source')
    data[time_col] = data[time_col].dt.tz_localize(None)

    # grouping and coloring sources by event "channels"
    data = data.join(colors[['channel']])\
        .sort_values(['channel', time_col], ascending=[False, True])

    if ax is None:
        n = data.index.unique().size
        fig, ax = plt.subplots(nrows=n, figsize=(10, n*height), sharex=True, squeeze=False)

    # plot data
    ax_shift = 0
    for i, (source, d) in enumerate(data.groupby('source', sort=False)):
        if (source not in CHANNELS) or (y_filter is not None and source not in y_filter):
            print(f'plot_channels: skipping {source}')
            ax_shift += 1
            continue
        iax = i - ax_shift
        if resample is not None:
            d = d.resample(resample, on=time_col).mean(numeric_only=True).reset_index()
        if source in colors.index:
            c = colors.loc[source, 'color']
        else:
            c = 'grey'
        if source in CHANNEL_LIMS:
            d = d.loc[(CHANNEL_LIMS[source][0] <= d['values']) & (d['values'] <= CHANNEL_LIMS[source][1])]
        ax[iax, 0].scatter(d[time_col].dt.tz_localize(None).values, d['values'].values, s=0.1, color=c)
        if source not in CHANNEL_LIMS:
            ylim = d['values'].quantile([0.001, 0.999]).tolist()
            ylim[0] = 0.95*ylim[0] if ylim[0] >= 0 else 1.1*ylim[0]
            ylim[1] = 1.1*ylim[1] if ylim[1] >= 0 else 0.95*ylim[1]
            ax[iax,0].set_ylim(*ylim)
        if source in rename_channels:
            ax[iax, 0].set_ylabel(rename_channels[source], rotation=0, horizontalalignment='right')
        else:
            ax[iax,0].set_ylabel(source, rotation=0, horizontalalignment='right')

        if source in discrete_events:
            if source in ENUMS:
                ax[iax, 0].set_yticks(ENUMS[source][0],labels=ENUMS[source][1])
            else:
                ax[iax, 0].set_yticks(d['values'].drop_duplicates().sort_values().values)
                ylabels = ax[iax, 0].get_yticklabels()
                for label in ylabels[1:-1]:
                    label.set_text('')
                ax[iax, 0].set_yticklabels(ylabels)

    # format plot
    for i in range(len(ax)):
        ax[i,0].set_xlabel('')
        ax[i,0].set_xticklabels([])
    if ax is None:
        print('entered')
        ax[-1,0].set_xlabel('Time')
        ax[-1,0].set_xlim(data[time_col].min(), data[time_col].max())
        format_xticks(ax[-1,0])


    return ax


def format_xticks(ax, format='%m/%d %H:%M'):
    """ format datestrings on x axis """
    xticks = ax.get_xticks()
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticks, rotation=25, ha='right')
    xfmt = mdates.DateFormatter(format)
    ax.xaxis.set_major_formatter(xfmt)


def get_legend_colors(cmap='muted'):
    # the following dict keys should correspond to sleep event channels
    colors = pd.Series(COLOR_GROUPS, name='source').to_frame().reset_index()\
        .rename(columns={'index': 'channel'}).sort_values('channel')
    colors['color'] = sns.color_palette(cmap, len(colors))

    return colors




# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/pheno-utils',
                'doc_host': 'https://pheno-ai.github.io',
                'git_url': 'https://github.com/pheno-ai/pheno-utils',
                'lib_path': 'pheno_utils'},
  'syms': { 'pheno_utils.age_reference_plots': { 'pheno_utils.age_reference_plots.AgeRefPlot': ( 'age_reference_plots.html#agerefplot',
                                                                                                 'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.AgeRefPlot.__init__': ( 'age_reference_plots.html#agerefplot.__init__',
                                                                                                          'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.AgeRefPlot.calc_linear_fit': ( 'age_reference_plots.html#agerefplot.calc_linear_fit',
                                                                                                                 'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.AgeRefPlot.plot': ( 'age_reference_plots.html#agerefplot.plot',
                                                                                                      'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.AgeRefPlot.plot_agehist': ( 'age_reference_plots.html#agerefplot.plot_agehist',
                                                                                                              'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.AgeRefPlot.plot_ornaments': ( 'age_reference_plots.html#agerefplot.plot_ornaments',
                                                                                                                'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.AgeRefPlot.plot_percentiles': ( 'age_reference_plots.html#agerefplot.plot_percentiles',
                                                                                                                  'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.AgeRefPlot.plot_scatter': ( 'age_reference_plots.html#agerefplot.plot_scatter',
                                                                                                              'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.AgeRefPlot.plot_valhist': ( 'age_reference_plots.html#agerefplot.plot_valhist',
                                                                                                              'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.GenderAgeRefPlot': ( 'age_reference_plots.html#genderagerefplot',
                                                                                                       'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.GenderAgeRefPlot.__init__': ( 'age_reference_plots.html#genderagerefplot.__init__',
                                                                                                                'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.GenderAgeRefPlot.plot': ( 'age_reference_plots.html#genderagerefplot.plot',
                                                                                                            'pheno_utils/age_reference_plots.py'),
                                                 'pheno_utils.age_reference_plots.get_gam_expectiles': ( 'age_reference_plots.html#get_gam_expectiles',
                                                                                                         'pheno_utils/age_reference_plots.py')},
            'pheno_utils.basic_analysis': { 'pheno_utils.basic_analysis.assign_nearest_research_stage': ( 'basic_analysis.html#assign_nearest_research_stage',
                                                                                                          'pheno_utils/basic_analysis.py'),
                                            'pheno_utils.basic_analysis.custom_describe': ( 'basic_analysis.html#custom_describe',
                                                                                            'pheno_utils/basic_analysis.py')},
            'pheno_utils.basic_plots': { 'pheno_utils.basic_plots.data_ecdfplot': ( 'basic_plots.html#data_ecdfplot',
                                                                                    'pheno_utils/basic_plots.py'),
                                         'pheno_utils.basic_plots.data_histplot': ( 'basic_plots.html#data_histplot',
                                                                                    'pheno_utils/basic_plots.py'),
                                         'pheno_utils.basic_plots.hist_ecdf_plots': ( 'basic_plots.html#hist_ecdf_plots',
                                                                                      'pheno_utils/basic_plots.py'),
                                         'pheno_utils.basic_plots.plot_data_collection': ( 'basic_plots.html#plot_data_collection',
                                                                                           'pheno_utils/basic_plots.py'),
                                         'pheno_utils.basic_plots.plot_hist_stats': ( 'basic_plots.html#plot_hist_stats',
                                                                                      'pheno_utils/basic_plots.py'),
                                         'pheno_utils.basic_plots.plot_stats': ( 'basic_plots.html#plot_stats',
                                                                                 'pheno_utils/basic_plots.py'),
                                         'pheno_utils.basic_plots.show_fundus': ( 'basic_plots.html#show_fundus',
                                                                                  'pheno_utils/basic_plots.py')},
            'pheno_utils.blandaltman_plots': { 'pheno_utils.blandaltman_plots.bland_altman_triple_plot': ( 'blandaltman_plots.html#bland_altman_triple_plot',
                                                                                                           'pheno_utils/blandaltman_plots.py')},
            'pheno_utils.cgm_plots': { 'pheno_utils.cgm_plots.AGP': ('cgm_plots.html#agp', 'pheno_utils/cgm_plots.py'),
                                       'pheno_utils.cgm_plots.AGP.__init__': ('cgm_plots.html#agp.__init__', 'pheno_utils/cgm_plots.py'),
                                       'pheno_utils.cgm_plots.AGP.plot': ('cgm_plots.html#agp.plot', 'pheno_utils/cgm_plots.py'),
                                       'pheno_utils.cgm_plots.CGMPlot': ('cgm_plots.html#cgmplot', 'pheno_utils/cgm_plots.py'),
                                       'pheno_utils.cgm_plots.CGMPlot.__init__': ( 'cgm_plots.html#cgmplot.__init__',
                                                                                   'pheno_utils/cgm_plots.py'),
                                       'pheno_utils.cgm_plots.CGMPlot.plot': ('cgm_plots.html#cgmplot.plot', 'pheno_utils/cgm_plots.py'),
                                       'pheno_utils.cgm_plots.CGMPlot.plot_diet': ( 'cgm_plots.html#cgmplot.plot_diet',
                                                                                    'pheno_utils/cgm_plots.py'),
                                       'pheno_utils.cgm_plots.CGMPlot.plot_gluc': ( 'cgm_plots.html#cgmplot.plot_gluc',
                                                                                    'pheno_utils/cgm_plots.py'),
                                       'pheno_utils.cgm_plots.CGMPlot.plot_sleep': ( 'cgm_plots.html#cgmplot.plot_sleep',
                                                                                     'pheno_utils/cgm_plots.py'),
                                       'pheno_utils.cgm_plots.CGMPlot.prep_plot': ( 'cgm_plots.html#cgmplot.prep_plot',
                                                                                    'pheno_utils/cgm_plots.py')},
            'pheno_utils.cohort_selector': { 'pheno_utils.cohort_selector.CohortSelector': ( 'cohort_selector.html#cohortselector',
                                                                                             'pheno_utils/cohort_selector.py'),
                                             'pheno_utils.cohort_selector.CohortSelector.__init__': ( 'cohort_selector.html#cohortselector.__init__',
                                                                                                      'pheno_utils/cohort_selector.py'),
                                             'pheno_utils.cohort_selector.CohortSelector.select': ( 'cohort_selector.html#cohortselector.select',
                                                                                                    'pheno_utils/cohort_selector.py')},
            'pheno_utils.config': { 'pheno_utils.config.copy_tre_config': ('config.html#copy_tre_config', 'pheno_utils/config.py'),
                                    'pheno_utils.config.generate_synthetic_data': ( 'config.html#generate_synthetic_data',
                                                                                    'pheno_utils/config.py'),
                                    'pheno_utils.config.generate_synthetic_data_like': ( 'config.html#generate_synthetic_data_like',
                                                                                         'pheno_utils/config.py')},
            'pheno_utils.config_setup.create_default_config': {},
            'pheno_utils.dates_plots': { 'pheno_utils.dates_plots.dates_dist_plot': ( 'date_plots.html#dates_dist_plot',
                                                                                      'pheno_utils/dates_plots.py')},
            'pheno_utils.ecg_analysis': { 'pheno_utils.ecg_analysis.get_hrv_df': ( 'ecg_analysis.html#get_hrv_df',
                                                                                   'pheno_utils/ecg_analysis.py'),
                                          'pheno_utils.ecg_analysis.vis_ecg': ('ecg_analysis.html#vis_ecg', 'pheno_utils/ecg_analysis.py')},
            'pheno_utils.meta_loader': { 'pheno_utils.meta_loader.MetaLoader': ( 'meta_loader.html#metaloader',
                                                                                 'pheno_utils/meta_loader.py'),
                                         'pheno_utils.meta_loader.MetaLoader.__concat__': ( 'meta_loader.html#metaloader.__concat__',
                                                                                            'pheno_utils/meta_loader.py'),
                                         'pheno_utils.meta_loader.MetaLoader.__get_dataset_path__': ( 'meta_loader.html#metaloader.__get_dataset_path__',
                                                                                                      'pheno_utils/meta_loader.py'),
                                         'pheno_utils.meta_loader.MetaLoader.__getitem__': ( 'meta_loader.html#metaloader.__getitem__',
                                                                                             'pheno_utils/meta_loader.py'),
                                         'pheno_utils.meta_loader.MetaLoader.__init__': ( 'meta_loader.html#metaloader.__init__',
                                                                                          'pheno_utils/meta_loader.py'),
                                         'pheno_utils.meta_loader.MetaLoader.__load_dictionaries__': ( 'meta_loader.html#metaloader.__load_dictionaries__',
                                                                                                       'pheno_utils/meta_loader.py'),
                                         'pheno_utils.meta_loader.MetaLoader.__repr__': ( 'meta_loader.html#metaloader.__repr__',
                                                                                          'pheno_utils/meta_loader.py'),
                                         'pheno_utils.meta_loader.MetaLoader.__str__': ( 'meta_loader.html#metaloader.__str__',
                                                                                         'pheno_utils/meta_loader.py'),
                                         'pheno_utils.meta_loader.MetaLoader.get': ( 'meta_loader.html#metaloader.get',
                                                                                     'pheno_utils/meta_loader.py'),
                                         'pheno_utils.meta_loader.MetaLoader.load': ( 'meta_loader.html#metaloader.load',
                                                                                      'pheno_utils/meta_loader.py')},
            'pheno_utils.pheno_loader': { 'pheno_utils.pheno_loader.PhenoLoader': ( 'pheno_loader.html#phenoloader',
                                                                                    'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__concat__': ( 'pheno_loader.html#phenoloader.__concat__',
                                                                                               'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__get_dataset_path__': ( 'pheno_loader.html#phenoloader.__get_dataset_path__',
                                                                                                         'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__get_dictionary_file_path__': ( 'pheno_loader.html#phenoloader.__get_dictionary_file_path__',
                                                                                                                 'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__get_file_path__': ( 'pheno_loader.html#phenoloader.__get_file_path__',
                                                                                                      'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__getitem__': ( 'pheno_loader.html#phenoloader.__getitem__',
                                                                                                'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__init__': ( 'pheno_loader.html#phenoloader.__init__',
                                                                                             'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__load_age_sex__': ( 'pheno_loader.html#phenoloader.__load_age_sex__',
                                                                                                     'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__load_dataframes__': ( 'pheno_loader.html#phenoloader.__load_dataframes__',
                                                                                                        'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__load_dictionary__': ( 'pheno_loader.html#phenoloader.__load_dictionary__',
                                                                                                        'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__load_one_dataframe__': ( 'pheno_loader.html#phenoloader.__load_one_dataframe__',
                                                                                                           'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__repr__': ( 'pheno_loader.html#phenoloader.__repr__',
                                                                                             'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.__str__': ( 'pheno_loader.html#phenoloader.__str__',
                                                                                            'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.describe_field': ( 'pheno_loader.html#phenoloader.describe_field',
                                                                                                   'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.get': ( 'pheno_loader.html#phenoloader.get',
                                                                                        'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.load_sample_data': ( 'pheno_loader.html#phenoloader.load_sample_data',
                                                                                                     'pheno_utils/pheno_loader.py'),
                                          'pheno_utils.pheno_loader.PhenoLoader.replace_bulk_data_path': ( 'pheno_loader.html#phenoloader.replace_bulk_data_path',
                                                                                                           'pheno_utils/pheno_loader.py')},
            'pheno_utils.questionnaires_handler': { 'pheno_utils.questionnaires_handler.convert_to_string': ( 'questionnaire_handler.html#convert_to_string',
                                                                                                              'pheno_utils/questionnaires_handler.py'),
                                                    'pheno_utils.questionnaires_handler.tranform_answers': ( 'questionnaire_handler.html#tranform_answers',
                                                                                                             'pheno_utils/questionnaires_handler.py')},
            'pheno_utils.sleep_plots': { 'pheno_utils.sleep_plots.format_xticks': ( 'sleep_plots.html#format_xticks',
                                                                                    'pheno_utils/sleep_plots.py'),
                                         'pheno_utils.sleep_plots.get_legend_colors': ( 'sleep_plots.html#get_legend_colors',
                                                                                        'pheno_utils/sleep_plots.py'),
                                         'pheno_utils.sleep_plots.plot_channels': ( 'sleep_plots.html#plot_channels',
                                                                                    'pheno_utils/sleep_plots.py'),
                                         'pheno_utils.sleep_plots.plot_events': ( 'sleep_plots.html#plot_events',
                                                                                  'pheno_utils/sleep_plots.py'),
                                         'pheno_utils.sleep_plots.plot_sleep': ( 'sleep_plots.html#plot_sleep',
                                                                                 'pheno_utils/sleep_plots.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_basic_analysis.ipynb.

# %% auto 0
__all__ = ['custom_describe', 'assign_nearest_research_stage']

# %% ../nbs/07_basic_analysis.ipynb 3
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Any, Dict, Union, Optional
import importlib

# %% ../nbs/07_basic_analysis.ipynb 4
try:
    importlib.import_module("smart_open")
except:
    print("smart_open not installed, please pip install smart-open")

from smart_open import open

# %% ../nbs/07_basic_analysis.ipynb 5
from .config import generate_synthetic_data

# %% ../nbs/07_basic_analysis.ipynb 6
def custom_describe(df: pd.DataFrame) -> pd.DataFrame:
    """
    Generates a custom summary statistics dataframe for mixed data types.
    
    Args:
        df: The input pandas DataFrame
        
    Returns:
        A pandas DataFrame containing the summary statistics
    """
    def describe_column(series: pd.Series) -> Dict[str, Union[int, float]]:
        """
        Generates summary statistics for a given column/series.
        
        Args:
            series: The input pandas Series (column)
            
        Returns:
            A dictionary containing the summary statistics
        """
        if pd.api.types.is_numeric_dtype(series):
            stats = {
                'count': series.count(),
                'unique': series.nunique(),
                'most_frequent': series.mode().iat[0] if series.mode().size > 0 else np.nan,
                'min': series.min(),
                'max': series.max(),
                'mean': series.mean(),
                'median': series.median(),
                'std': series.std(),
            }
        elif pd.api.types.is_datetime64_dtype(series):
            stats = {
                'count': series.count(),
                'unique': series.nunique(),
                'most_frequent': np.nan,
                'min': series.min(),
                'max': series.max(),
                'mean': np.nan,
                'median': np.nan,
                'std': np.nan,
            }
        else:
            stats = {
                'count': series.count(),
                'unique': series.nunique(),
                'most_frequent': series.mode().iat[0] if series.mode().size > 0 else np.nan,
                'min': np.nan,
                'max': np.nan,
                'mean': np.nan,
                'median': np.nan,
                'std': np.nan,
            }
        return stats

    summary = pd.DataFrame({col: describe_column(df[col]) for col in df.columns}).transpose()
    summary = summary[['count', 'unique', 'most_frequent', 'min', 'max', 'mean', 'median', 'std']]
    return summary.T

# %% ../nbs/07_basic_analysis.ipynb 8
def assign_nearest_research_stage(dataset: pd.DataFrame, 
                                  population: pd.DataFrame, 
                                  max_days: int = 60, 
                                  stages: List[str] = ['visit'], 
                                  agg: Union[str, None] = 'first') -> pd.DataFrame:
    """
    Assign the nearest research stage to each record in a dataset.
    
    Args:
        dataset (pd.DataFrame): The dataset containing records to be assigned research stages.
        population (pd.DataFrame): The population data with participant_id, cohort, research_stage, and research_stage_date.
        max_days (int, optional): The maximum number of days allowed between the collection date and research stage date. Defaults to 60.
        stages (List[str], optional): The list of types of research stages to consider. Defaults to ['visit'].
        agg (Union[str, None], optional): The aggregation function to be used when (optionally) aggregating multiple rows
                                          from the same research stage. The rows are already sorted by distance from the
                                          date of the research stage. Can be 'first' (closest), 'last' (farthest), 'mean',
                                          'min', 'max', or None. Defaults to 'first'.

    Returns:
        pd.DataFrame: The dataset with the nearest research stage assigned to each record.
    """
    data_wo_stage = dataset.sort_values('collection_date').reset_index()\
        .drop(columns=['research_stage'], errors='ignore')
    population = population.loc[
        population.index.get_level_values('research_stage').str.contains('|'.join(stages))]\
        .sort_values('research_stage_date').reset_index()\
        [['participant_id', 'cohort', 'research_stage', 'research_stage_date']]

    data_w_stage = pd.merge_asof(
        data_wo_stage,
        population,
        by=['participant_id', 'cohort'],
        left_on='collection_date',
        right_on='research_stage_date',
        direction='nearest')\
        .assign(delta=lambda x: (x['collection_date'] - x['research_stage_date']).dt.days.abs())
    data_w_stage = data_w_stage.loc[data_w_stage['delta'] < max_days].sort_values('delta')\
        .drop(columns=['delta', 'research_stage_date'])

    if agg is None:
        # return data_wo_stage, data_w_stage
        return data_wo_stage.merge(
            data_w_stage[['participant_id', 'cohort', 'collection_date', 'research_stage']].drop_duplicates(),
            how='left')\
            .set_index(dataset.index.names)

    return data_w_stage.groupby(['participant_id', 'cohort', 'research_stage']).agg(agg)\
        .drop(columns=['array_index'], errors='ignore')




import os 
import json
import argparse

base_json = {
"DATASETS_PATH": "",
"BULK_DATA_PATH": {
    "\\./": ""
},
"EVENTS_DATASET": "events",
"COHORT": "None",
"ERROR_ACTION": "warn"
}

bulk_data_mapping = {
"s3://.*/fundus/10k/": "{bulk_data_path}/fundus_bulk/", 
"s3://.*/human_genetics/10k/": "{bulk_data_path}/human_genetics_bulk/", 
"s3://.*/gut_microbiome/10k/": "{bulk_data_path}/gut_microbiome_bulk/"
}

TRE_BASE_PATH = '/home/ec2-user/studies/'


def handle_arguments():
    argparser = argparse.ArgumentParser(description='Create config.json file for pheno')
    argparser.add_argument('-d', '--datasets_path', type=str, help='datasets path on s3 or tre')
    argparser.add_argument('-events', '--events_dataset', default="events", type=str, help='events dataset name')
    argparser.add_argument('-cohort', '--cohort', default="None", type=str, help='cohort name')
    argparser.add_argument('-error', '--error_action', default="warn", type=str, help='error action')
    argparser.add_argument('-b', '--bulk_data_path', type=str, default=None, help='bulk data path')
    argparser.add_argument('-m', '--mode', type=str, default=None, help='enter "tre" if thats the case')
    argparser.add_argument('-c', '--client', type=str, default=None, help='client name')
    
    
    return argparser.parse_args()


def copy_tre_config():
    default_config_found = False
    script_path = os.path.dirname(os.path.abspath(__file__))
    
    env_config = os.path.join(script_path, 'env_config.json')
    with open(env_config, 'r') as openfile:
        env_json = json.load(openfile)
    
    # where am I? 
    config_name = ''
    env = None
    for k, v in env_json.items():
        if os.path.exists(v['ident_path']):
            env = k
            config_name = v['config_name']
            break
        
    absolute_config_path = os.path.join(script_path, config_name)
    # with open(absolute_config_path, 'r') as openfile:
    #     json_object = json.load(openfile)
                
    # datasets_full_path = json_object['DATASETS_PATH']
    if (env is not None) and (os.path.exists(absolute_config_path)):
        default_config_found = True
        if not os.path.exists(os.path.expanduser('~/.pheno')):
            os.makedirs(os.path.expanduser('~/.pheno'))
        
        shutil.copy2(absolute_config_path, os.path.expanduser('~/.pheno/config.json'))
    
    return default_config_found



def main():
    args = handle_arguments()
    
    if not copy_tre_config(): 

        print('S3 or Local Mode')
        if args.datasets_path is None: 
            print('Please enter a datasets path using -d or --datasets_path parameter')
            return
        
        base_json["DATASETS_PATH"] = args.datasets_path
        base_json["EVENTS_DATASET"] = args.events_dataset
        base_json["COHORT"] = args.cohort
        base_json["ERROR_ACTION"] = args.error_action
        base_json["BULK_DATA_PATH"]["\\./"] = args.datasets_path + "/{dataset}/"
        
        if args.bulk_data_path:
            for k, v in bulk_data_mapping.items():
                base_json["BULK_DATA_PATH"][k] = v.format(bulk_data_path=args.bulk_data_path)
        
        
        if not os.path.exists(os.path.expanduser('~/.pheno')):
            os.makedirs(os.path.expanduser('~/.pheno'))
        
        with open(os.path.expanduser('~/.pheno/config.json'), 'w') as outfile:
            json.dump(base_json, outfile, indent=4)


# run main 
if __name__ == '__main__':
    main()


# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_blandaltman_plots.ipynb.

# %% auto 0
__all__ = ['bland_altman_triple_plot']

# %% ../nbs/02_blandaltman_plots.ipynb 3
from .config import *

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.stats import pearsonr
from typing import Tuple
import importlib

# %% ../nbs/02_blandaltman_plots.ipynb 4
try:
    importlib.import_module("pyCompare")
except ImportError:
    raise ImportError("The 'pyCompare' library is not installed. Please install it using 'pip install pyCompare' before running this script.")

from pyCompare import blandAltman

# %% ../nbs/02_blandaltman_plots.ipynb 5
def bland_altman_triple_plot(
    data: pd.DataFrame, m1_col: str, m2_col: str, feature_str: str = "",
    scale: str = "linear",
) -> None:
    """
    Generates a triple plot consisting of a scatter correlation plot, Bland-Altman plot, and a percentage Bland-Altman plot.

    Args:
        data (pd.DataFrame): A pandas DataFrame containing the data.
        m1_col (str): The name of the first measurement column in the DataFrame.
        m2_col (str): The name of the second measurement column in the DataFrame.
        feature_str (str, optional): A string to include in the title of the plots. Defaults to "".
        scale (str, optional): The scale of the axes. Defaults to "linear".

    Returns:
        None
    """

    m1 = data.dropna(subset=[m1_col, m2_col])[m1_col] 
    m2 = data.dropna(subset=[m1_col, m2_col])[m2_col]
    min_val = np.min([np.min(m1), np.min(m2)])
    max_val = np.min([np.max(m1), np.max(m2)])

    fig, axes = plt.subplots(1, 3, figsize=(18, 5))
    fig.suptitle(f"{feature_str} - Bland-Altman plots", fontsize=16)

    ax = axes[0]
    sns.regplot(x=m1, y=m2, ax=ax, scatter_kws=dict(alpha=0.3))
    ax.plot(
        [min_val * 0.99, max_val * 1.01],
        [min_val * 0.99, max_val * 1.01],
        ls=":",
        color="grey",
    )
    ax.set_xlim(min_val * 0.99, max_val * 1.01)
    ax.set_ylim(min_val * 0.99, max_val * 1.01)
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    r = np.round(pearsonr(data.dropna(subset=[m1_col,m2_col])[m1_col], data.dropna(subset=[m1_col,m2_col])[m2_col])[0], 3)
    ax.set_title(f"r={r}")

    ax = axes[1]
    blandAltman(m1, m2, ax=ax)
    ax.set_xlabel(f"Mean of {m1.name} and {m2.name}")
    ax.set_ylabel(f"{m1.name} - {m2.name}")
    ax.set_xscale(scale)

    ax = axes[2]
    blandAltman(m1, m2, ax=ax, percentage=True)
    ax.set_xlabel(f"Mean of {m1.name} and {m2.name}")
    ax.set_ylabel(f"Percentage ({m1.name} - {m2.name})")
    ax.set_xscale(scale)




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/11_meta_loader.ipynb.

# %% auto 0
__all__ = ['MetaLoader']

# %% ../nbs/11_meta_loader.ipynb 3
import os
import re
from typing import List, Any, Dict, Union
import warnings

import numpy as np
import pandas as pd
import dask.dataframe as dd

# %% ../nbs/11_meta_loader.ipynb 4
from pheno_utils.config import (
    DATASETS_PATH, 
    COHORT, 
    ERROR_ACTION
    )
from .pheno_loader import PhenoLoader

# %% ../nbs/11_meta_loader.ipynb 5
class MetaLoader:
    """
    Class to load multiple dictionaries and allows to easily access the relevant fields.

    Args:
    
        base_path (str, optional): The base path where the data is stored. Defaults to DATASETS_PATH.
        cohort (str, optional): The name of the cohort within the dataset. Defaults to COHORT.
        flexible_field_search (bool, optional): Whether to allow regex field search. Defaults to False.
        errors (str, optional): Whether to raise an error or issue a warning if missing data is encountered.
            Possible values are 'raise', 'warn' and 'ignore'. Defaults to 'raise'.
        **kwargs: Additional keyword arguments to pass to a DataLoader class.

    Attributes:
    
        dicts (pd.DataFrame): A dictionary of data dictionaries (dataframes) of all availbale datasets in the base_path.
        fields (list): A list of all fields.
        cohort (str): The name of the cohort being used.
        base_path (str): The base path where the data is stored.
        flexible_field_search (bool): Whether to allow regex field search.
        errors (str): Whether to raise an error or issue a warning if missing data is encountered.
        kwargs (dict): Additional keyword arguments to pass to a DataLoader class.
    """

    def __init__(
        self,
        base_path: str = DATASETS_PATH,
        cohort: str = COHORT,
        flexible_field_search: bool = False,
        errors: str = ERROR_ACTION,
        **kwargs,
    ) -> None:
        self.cohort = cohort
        self.base_path = base_path
        self.dataset_path = self.__get_dataset_path__()
        
        self.flexible_field_search = flexible_field_search
        self.errors = errors
        self.kwargs = kwargs

        self.__load_dictionaries__()

    def load(self, fields: Union[str,List[str]], flexible: bool=None, prop: str='tabular_field_name') -> pd.DataFrame:
        """
        Return a dataframe containing the fields from the respective datasets.

        Args:
            fields (Union[str,List[str]]): Fields to return
            flexible (bool, optional): Whether to use fuzzy matching to find fields. Defaults to None, which uses the DataLoader's flexible_field_search attribute.
            prop (str, optional): The property to use for searching. Defaults to 'tabular_field_name'.

        Returns:
            pd.DataFrame: Dataframe containing the fields from the respective datasets.
        """
        found_fields = self.get(fields, flexible, prop)
        found_fields.columns = found_fields.columns.str.split('/').str[1]
        dup_fields = found_fields.columns.value_counts()\
            .to_frame('count').query('count > 1').index

        loaded_fields = []
        for ds, f in found_fields.T.groupby('dataset'):
            df = PhenoLoader(ds, base_path=self.base_path, cohort=self.cohort,
                             age_sex_dataset=None, **self.kwargs)\
                [f.index.tolist()]
            # rename duplicate fields
            df = df.rename(columns=pd.Series(dup_fields + f'_{ds}', index=dup_fields))

            if not len(loaded_fields):
                loaded_fields = df
                continue

            loaded_fields = loaded_fields.join(df, how='outer')

        return loaded_fields

    def get(self, fields: Union[str,List[str]], flexible: bool=None, prop='tabular_field_name') -> pd.DataFrame:
        """
        Return metadata for the specified fields from all tables.

        Args:
            fields (List[str]): Fields to return
            flexible (bool, optional): Whether to use fuzzy matching to find fields. Defaults to None, which uses the DataLoader's flexible_field_search attribute.
            prop (str, optional): The property to use for searching. Defaults to 'tabular_field_name'.

        Returns:
            pd.DataFrame: Data for the specified fields from all tables
        """
        if flexible is None:
            flexible = self.flexible_field_search
        if isinstance(fields, str):
            fields = [fields]
        fields = [f.lower() for f in fields]

        data = pd.DataFrame()
        for dataset, df in self.dicts.items():
            if prop == 'tabular_field_name':
                search_in = pd.Series(df.columns, index=df.columns).str.lower()
            else:
                search_in = df.loc[prop].dropna().str.lower()
            if flexible:
                # use fuzzy matching including regex to find fields
                fields_in_col = np.unique([col for f in fields for col, text in search_in.items()
                                           if type(text) is str and re.search(f, text)])
            else:
                fields_in_col = search_in[search_in.isin(fields)].index
            if len(fields_in_col):
                this_data = df[fields_in_col]
                this_data.columns = dataset + '/' + this_data.columns
                data = self.__concat__(data, this_data)

        return data

    def __repr__(self):
        """
        Return string representation of object

        Returns:
            str: String representation of object
        """
        return self.__str__()

    def __str__(self):
        """
        Return string representation of object

        Returns:
            str: String representation of object
        """
        ds_list = str(list(self.dicts.keys())).replace(',', '\n')
        return f'MetaLoader for: {self.dataset_path}\n' + \
            f'with {len(self.fields)} fields\n{len(self.dicts)} datasets:\n{ds_list}'

    def __getitem__(self, fields: Union[str,List[str]]):
        """
        Return data for the specified fields from all tables

        Args:
            fields (Union[str, List[str]]): Fields to return

        Returns:
            pd.DataFrame: Data for the specified fields from all tables
        """
        return self.get(fields)

    def __concat__(self, df1, df2):
        if df1.empty:
            return df2
        if df2.empty:
            return df1
        return df1.join(df2, how='outer')

    def __load_dictionaries__(self) -> None:
        """
        Load all dictionaries in the base_path.
        """
        dicts = dd.read_csv(os.path.join(self.dataset_path, '*_dict*.csv'),
                            include_path_column=True, dtype={'parent_dataframe': 'object', 'sampling_rate': 'object'}).compute()
        if self.cohort is None:
            dataset_ind = -2
        else:
            dataset_ind = -3
        dicts['dataset'] = dicts['path'].str.split('/').str[dataset_ind]
        dicts = dicts.drop(columns=['path'])
        self.fields = dicts['tabular_field_name'].unique()

        self.dicts = {}
        col_order = ['dataset'] + dicts.columns.drop('dataset').tolist()
        for dataset in dicts['dataset'].unique():
            self.dicts[dataset] = dicts.loc[dicts['dataset'] == dataset, col_order].set_index('tabular_field_name').T

    def __get_dataset_path__(self):
        """
        Get the dataset path.

        Args:
            dataset (str): the name of the dataset

        Returns:
            str: the path to the dataset
        """
        if self.cohort is not None:
            return os.path.join(self.base_path, '*', self.cohort)
        return os.path.join(self.base_path, '*')




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/08_cgm_plots.ipynb.

# %% auto 0
__all__ = ['GLUC_COLOR', 'DIET_COLOR', 'SLEEP_COLOR', 'FONTSIZE', 'CGMPlot', 'AGP']

# %% ../nbs/08_cgm_plots.ipynb 3
from typing import List, Tuple, Optional

import numpy as np
import pandas as pd

import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

# %% ../nbs/08_cgm_plots.ipynb 4
GLUC_COLOR = "C0"
DIET_COLOR = "brown"
SLEEP_COLOR = "purple"
FONTSIZE = 14


class CGMPlot:
    def __init__(
        self,
        cgm_df: pd.DataFrame,
        diet_df: Optional[pd.DataFrame] = None,
        cgm_date_col: str = "collection_timestamp",
        gluc_col: str = "glucose",
        diet_date_col: str = "collection_timestamp",
        diet_text_col: str = "shortname_eng",
        ax: Optional[plt.Axes] = None,
        smooth: bool = False,
        sleep_tuples: Optional[List[Tuple[pd.Timestamp, pd.Timestamp]]] = None,
    ) -> None:
        """
        Initialize a CGMPlot object.

        Args:
            cgm_df (pd.DataFrame): DataFrame containing the glucose measurements.
            diet_df (Optional[pd.DataFrame], optional): DataFrame containing the diet data. Defaults to None.
            cgm_date_col (str, optional): Name of the date column in cgm_df. Defaults to "Date".
            gluc_col (str, optional): Name of the glucose column in cgm_df. Defaults to "glucose".
            diet_date_col (str, optional): Name of the date column in diet_df. Defaults to "Date".
            diet_text_col (str, optional): Name of the text column in diet_df. Defaults to "shortname_eng".
            ax (Optional[plt.Axes], optional): Matplotlib Axes object to plot on. Defaults to None.
            smooth (bool, optional): Apply smoothing to the glucose curve. Defaults to False.
            sleep_tuples (Optional[List[Tuple[pd.Timestamp, pd.Timestamp]]], optional): List of sleep start and end times. Defaults to None.
        """
        self.cgm_df = cgm_df
        self.diet_df = diet_df
        self.cgm_date_col = cgm_date_col
        self.gluc_col = gluc_col
        self.diet_date_col = diet_date_col
        self.diet_text_col = diet_text_col
        self.smooth = smooth
        self.n_points = len(self.cgm_df[self.gluc_col])
        self.gluc_color = GLUC_COLOR
        self.diet_color = DIET_COLOR
        self.datetime_start = self.cgm_df[self.cgm_date_col].iloc[0]
        self.sleep_tuples = sleep_tuples

        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(18, 5))
        self.ax = ax

    def prep_plot(self) -> None:
        """
        Prepare the plot by hiding the right and top spines.
        """
        self.ax.spines["right"].set_visible(False)
        self.ax.spines["top"].set_visible(False)

    def plot_gluc(self) -> None:
        """
        Plot the glucose data.
        """
        ax = self.ax
        y = self.cgm_df[self.gluc_col]
        x = self.cgm_df[self.cgm_date_col]
        if self.smooth:
            # smoothing
            smoothed = (
                self.cgm_df.set_index(self.cgm_date_col)
                .asfreq(freq="60S")
                .interpolate(method="cubicspline")
            )
            ax.plot(
                smoothed.index,
                smoothed[self.gluc_col],
                ls="-",
                lw=4,
                color=self.gluc_color,
                alpha=0.8,
                )
        else:
            ax.plot(x, y, ls="-", lw=4, color=self.gluc_color, alpha=0.9)
        ax.scatter(x, y, s=60, color=self.gluc_color, alpha=0.6, label="Glucose")

    def plot_diet(self) -> None:
        """
        Plot the diet data.
        """
        ax = self.ax
        x_offset = pd.to_timedelta(-5, "minutes")
        max_y = ax.get_ylim()[1]
        min_y = ax.get_ylim()[0]
        ax.set_ylim(min(60, min_y), max(150, max_y))

        for i, (food_datetime, group) in enumerate(
            self.diet_df.groupby(self.diet_date_col)
        ):
            food_str = "\n".join(group[self.diet_text_col])

            txt_x = food_datetime - pd.to_timedelta(7.5, "m")
            if i % 2 == 0:
                txt_y = max(150, max_y) + 1
                horizontalalignment = "center"
            else:
                txt_y = max(150, max_y) + 5
                horizontalalignment = "center"

            ax.axvline(food_datetime, color=self.diet_color, alpha=0.5, ls=":")
            ax.scatter(
                x=food_datetime,
                y=max(150, max_y) - 2,
                marker="v",
                s=100,
                color=self.diet_color,
            )
            ax.text(
                x=txt_x,
                y=txt_y,
                s=(food_str),
                color=self.diet_color,
                horizontalalignment=horizontalalignment,
                fontsize=10,
            )

    def plot_sleep(self) -> None:
        """
        Plot the sleep data.
        """
        ax = self.ax
        for sleep_start, sleep_end in self.sleep_tuples:
            ax.fill_between(
                [sleep_start, sleep_end],
                ax.get_ylim()[0],
                ax.get_ylim()[1],
                color=SLEEP_COLOR,
                alpha=0.05,
            )

    def plot(self) -> None:
        """
        Plot the CGMPlot object.
        """
        self.prep_plot()
        self.plot_gluc()
        if self.diet_df is not None:
            self.plot_diet()
        if self.sleep_tuples is not None:
            self.plot_sleep()


# %% ../nbs/08_cgm_plots.ipynb 10
class AGP:
    def __init__(
        self,
        cgm_df: pd.DataFrame,
        cgm_date_col: str = "collection_timestamp",
        gluc_col: str = "glucose",
        ax: Optional[plt.Axes] = None,
    ) -> None:
        """
        Initialize an AGP object.

        Args:
            cgm_df (pd.DataFrame): DataFrame containing the glucose measurements.
            cgm_date_col (str, optional): Name of the date column in cgm_df. Defaults to "collection_timestamp".
            gluc_col (str, optional): Name of the glucose column in cgm_df. Defaults to "glucose".
            ax (Optional[plt.Axes], optional): Matplotlib Axes object to plot on. Defaults to None.
        """
        self.cgm_df = cgm_df
        self.cgm_date_col = cgm_date_col
        self.gluc_col = gluc_col
        self.gluc_color = GLUC_COLOR

        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(18, 5))
        self.ax = ax

        agp_df = (
            cgm_df.set_index(cgm_date_col)
            .asfreq(freq="60S")
            .interpolate(method="cubicspline")
            .reset_index()
        )
        agp_df["minute_in_day"] = 60 * agp_df[self.cgm_date_col].dt.hour + agp_df[self.cgm_date_col].dt.minute
        self.agp_df = agp_df

    def plot(self) -> None:
        """
        Plot the AGP object.
        """
        ax = self.ax
        agp_df = self.agp_df
        median = agp_df.groupby("minute_in_day")["glucose"].median()
        lo_5 = agp_df.groupby("minute_in_day")["glucose"].quantile(0.05)
        hi_95 = agp_df.groupby("minute_in_day")["glucose"].quantile(0.95)
        lo_25 = agp_df.groupby("minute_in_day")["glucose"].quantile(0.25)
        hi_75 = agp_df.groupby("minute_in_day")["glucose"].quantile(0.75)

        ax.plot(median, color="k", lw=3)
        ax.fill_between(median.index.values, lo_25, hi_75, color="navy", alpha=0.3)
        ax.fill_between(median.index.values, lo_5, hi_95, color="navy", alpha=0.1)

        xticks = [m for m in median.index.values if m % 180 == 0]
        xticks += [24 * 60]
        xticklabels = [f"{int(m/60)}:00" for m in xticks]
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels)

        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.set_ylabel("Glucose (mg\dL)", fontsize=14)



__version__ = "0.3.8.10"

from .age_reference_plots import *
from .basic_analysis import *
from .basic_plots import *
from .blandaltman_plots import *
from .cgm_plots import *
from .config import *
from .pheno_loader import PhenoLoader
from .dates_plots import *
from .ecg_analysis import *
from .sleep_plots import *
from .meta_loader import *
